from abc import ABC, abstractmethod
import argparse
import inspect
import logging
import os
import pathlib
import sys
from typing import Optional, Sequence

from platypush.config import Config
from platypush.utils.manifest import (
    Dependencies,
    InstallContext,
)

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger()


class BaseBuilder(ABC):
    """
    Base interface and utility methods for Platypush builders.

    A Platypush builder is a script/piece of logic that can build a Platypush
    installation, with all its base and required extra dependencies, given a
    configuration file.

    This class is currently implemented by the :module:`platypush.platyvenv`
    and :module:`platypush.platydock` modules/scripts.
    """

    REPO_URL: str = 'https://github.com/BlackLight/platypush.git'
    """
    We use the Github URL here rather than the self-hosted Gitea URL to prevent
    too many requests to the Gitea server.
    """

    def __init__(
        self,
        cfgfile: str,
        gitref: str,
        output: str,
        install_context: InstallContext,
        *_,
        verbose: bool = False,
        device_id: Optional[str] = None,
        **__,
    ) -> None:
        """
        :param cfgfile: The path to the configuration file.
        :param gitref: The git reference to use. It can be a branch name, a tag
            name or a commit hash.
        :param output: The path to the output file or directory.
        :param install_context: The installation context for this builder.
        :param verbose: Whether to log debug traces.
        :param device_id: A device name that will be used to uniquely identify
            this installation.
        """
        self.cfgfile = os.path.abspath(os.path.expanduser(cfgfile))
        self.output = os.path.abspath(os.path.expanduser(output))
        self.gitref = gitref
        self.install_context = install_context
        self.device_id = device_id
        logger.setLevel(logging.DEBUG if verbose else logging.INFO)

    @classmethod
    @abstractmethod
    def get_name(cls) -> str:
        """
        :return: The name of the builder.
        """

    @classmethod
    @abstractmethod
    def get_description(cls) -> str:
        """
        :return: The description of the builder.
        """

    @property
    def deps(self) -> Dependencies:
        """
        :return: The dependencies for this builder, given the configuration
            file and the installation context.
        """
        return Dependencies.from_config(
            self.cfgfile,
            install_context=self.install_context,
        )

    def _print_instructions(self, s: str) -> None:
        GREEN = '\033[92m'
        NORM = '\033[0m'

        helper_lines = s.split('\n')
        wrapper_line = '=' * max(len(t) for t in helper_lines)
        helper = '\n' + '\n'.join([wrapper_line, *helper_lines, wrapper_line]) + '\n'
        print(GREEN + helper + NORM)

    @abstractmethod
    def build(self):
        """
        Builds the application. To be implemented by the subclasses.
        """

    @classmethod
    def _get_arg_parser(cls) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            prog=cls.get_name(),
            add_help=False,
            description=cls.get_description(),
        )

        parser.add_argument(
            '-h', '--help', dest='show_usage', action='store_true', help='Show usage'
        )

        parser.add_argument(
            '-v',
            '--verbose',
            dest='verbose',
            action='store_true',
            help='Enable debug traces',
        )

        parser.add_argument(
            '-c',
            '--config',
            type=str,
            dest='cfgfile',
            required=False,
            default=None,
            help='The path to the configuration file. If not specified, a minimal '
            'installation including only the base dependencies will be generated.',
        )

        parser.add_argument(
            '-o',
            '--output',
            dest='output',
            type=str,
            required=False,
            default='.',
            help='Target directory (default: current directory). For Platydock, '
            'this is the directory where the Dockerfile will be generated. For '
            'Platyvenv, this is the base directory of the new virtual '
            'environment.',
        )

        parser.add_argument(
            '-d',
            '--device-id',
            dest='device_id',
            type=str,
            required=False,
            default=None,
            help='A name that will be used to uniquely identify this device. '
            'Default: a random name for Docker containers, and the '
            'hostname of the machine for virtual environments.',
        )

        parser.add_argument(
            '-r',
            '--ref',
            dest='gitref',
            required=False,
            type=str,
            default='master',
            help='If the script is not run from a Platypush installation directory, '
            'it will clone the sources via git. You can specify through this '
            'option which branch, tag or commit hash to use. Defaults to master.',
        )

        return parser

    @classmethod
    def from_cmdline(cls, args: Sequence[str]):
        """
        Create a builder instance from command line arguments.

        :param args: Command line arguments.
        :return: A builder instance.
        """
        parser = cls._get_arg_parser()
        opts, _ = parser.parse_known_args(args)
        if opts.show_usage:
            parser.print_help()
            sys.exit(0)

        if not opts.cfgfile:
            opts.cfgfile = os.path.join(
                str(pathlib.Path(inspect.getfile(Config)).parent),
                'config.yaml',
            )

            logger.info('No configuration file specified. Using %s.', opts.cfgfile)

        return cls(**vars(opts))



from collections import namedtuple
from dataclasses import dataclass, field
from threading import Event, RLock
from typing import Optional

AudioFrame = namedtuple('AudioFrame', ['data', 'timestamp'])


@dataclass
class PauseState:
    """
    Data class to hold the boilerplate (state + synchronization events) for the
    audio recorder pause API.
    """

    _paused_event: Event = field(default_factory=Event)
    _recording_event: Event = field(default_factory=Event)
    _state_lock: RLock = field(default_factory=RLock)

    @property
    def paused(self):
        with self._state_lock:
            return self._paused_event.is_set()

    def pause(self):
        """
        Pause the audio recorder.
        """
        with self._state_lock:
            self._paused_event.set()
            self._recording_event.clear()

    def resume(self):
        """
        Resume the audio recorder.
        """
        with self._state_lock:
            self._paused_event.clear()
            self._recording_event.set()

    def toggle(self):
        """
        Toggle the audio recorder pause state.
        """
        with self._state_lock:
            if self.paused:
                self.resume()
            else:
                self.pause()

    def wait_paused(self, timeout: Optional[float] = None):
        """
        Wait until the audio recorder is paused.
        """
        self._paused_event.wait(timeout=timeout)

    def wait_recording(self, timeout: Optional[float] = None):
        """
        Wait until the audio recorder is resumed.
        """
        self._recording_event.wait(timeout=timeout)



from enum import Enum


class XmppPresence(Enum):
    """
    Models the XMPP presence states.
    """

    AVAILABLE = "available"
    NONE = AVAILABLE
    PLAIN = AVAILABLE
    OFFLINE = "offline"
    XA = "xa"
    EXTENDED_AWAY = XA
    AWAY = "away"
    CHAT = "chat"
    FREE_FOR_CHAT = CHAT
    DND = "dnd"
    DO_NOT_DISTURB = DND



from ._base import AudioConverter
from ._from_raw import RawInputAudioConverter
from ._to_raw import RawOutputAudioConverter, RawOutputAudioFromFileConverter

__all__ = [
    'AudioConverter',
    'RawInputAudioConverter',
    'RawOutputAudioConverter',
    'RawOutputAudioFromFileConverter',
]



from uuid import UUID

from sqlalchemy import (
    Boolean,
    Column,
    ForeignKey,
    Integer,
    String,
)

from platypush.common.db import is_defined
from platypush.entities import Entity

if not is_defined('bluetooth_service'):

    class BluetoothService(Entity):
        """
        Entity that represents a Bluetooth service.
        """

        __tablename__ = 'bluetooth_service'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        _uuid = Column('uuid', String, nullable=False)
        """
        The service class UUID. It can be either a 16-bit or a 128-bit UUID.
        """

        _protocol = Column('protocol', String, default=None)
        """ The protocol used by the service. """

        port = Column(Integer, default=None)
        """ The port used by the service. """

        version = Column(Integer, default=None)
        """ The version of the service profile. """

        is_ble = Column(Boolean, default=False)
        """ Whether the service is a BLE service. """

        connected = Column(Boolean, default=False)
        """ Whether an active connection exists to this service. """

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }

        @staticmethod
        def to_uuid(value):
            """
            Convert a raw UUID string to a service class UUID.
            """
            # If it's already a UUID or an int, just return it.
            if isinstance(value, (UUID, int)):
                return value
            try:
                # If it's formatted like a 128-bit UUID, convert it to a UUID
                # object.
                return UUID(value)
            except ValueError:
                # Hex string case
                return int(value, 16)

        @property
        def uuid(self):
            """
            Getter for the service class UUID.
            """
            return self.to_uuid(self._uuid)

        @uuid.setter
        def uuid(self, value):
            """
            Setter for the service class UUID.
            """
            uuid = self.to_uuid(value)
            if isinstance(uuid, int):
                # Hex-encoded 16-bit UUID case
                uuid = f'{uuid:04X}'

            self._uuid = str(uuid)

        @property
        def protocol(self):
            """
            Getter for the protocol used by the service.
            """
            from platypush.plugins.bluetooth.model import Protocol

            try:
                return Protocol(self._protocol)
            except ValueError:
                return Protocol.UNKNOWN

        @protocol.setter
        def protocol(self, value):
            """
            Setter for the protocol used by the service.
            """
            from platypush.plugins.bluetooth.model import Protocol

            protocol = Protocol.UNKNOWN
            if isinstance(value, Protocol):
                protocol = value
            else:
                try:
                    protocol = Protocol(value)
                except ValueError:
                    pass

            self._protocol = protocol.value

        @property
        def service_class(self):
            """
            The :class:`platypush.plugins.bluetooth.model.ServiceClass` enum
            value.
            """
            from platypush.plugins.bluetooth.model import ServiceClass

            try:
                return ServiceClass.get(self.uuid)
            except (TypeError, ValueError):
                return ServiceClass.UNKNOWN

        def to_dict(self) -> dict:
            return {
                **{k.lstrip('_'): v for k, v in super().to_dict().items()},
                # Human-readable service class name
                'service_class': str(self.service_class),
            }



from enum import Enum


class AssistantState(Enum):
    """
    Possible states of the assistant.
    """

    IDLE = 'idle'
    DETECTING_HOTWORD = 'detecting_hotword'
    DETECTING_SPEECH = 'detecting_speech'


# vim:sw=4:ts=4:et:



from typing import Iterable

from ._base import AudioConverter, dtype_to_ffmpeg_format


class RawOutputAudioConverter(AudioConverter):
    """
    Converts input audio to raw audio output.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._dtype and not self._output_format:
            ffmpeg_format = dtype_to_ffmpeg_format.get(self._dtype)
            assert ffmpeg_format, (
                f'Unsupported data type: {self._dtype}. Supported data types: '
                f'{list(dtype_to_ffmpeg_format.keys())}'
            )

            self._output_format = ffmpeg_format

    @property
    def _input_format_args(self) -> Iterable[str]:
        if not self._input_format:
            return ()

        ffmpeg_args = self._format_to_ffmpeg_args.get(self._input_format)
        if not ffmpeg_args:
            return ('-f', self._input_format)

        return ffmpeg_args

    @property
    def _output_format_args(self) -> Iterable[str]:
        args = (
            '-ar',
            str(self._sample_rate),
            *self._channel_layout_args,
        )

        if self._output_format:
            args = ('-f', self._output_format) + args

        return args


class RawOutputAudioFromFileConverter(RawOutputAudioConverter):
    """
    Converts an input file to raw audio output.
    """

    def __init__(self, *args, infile: str, **kwargs):
        super().__init__(*args, **kwargs)
        self.infile = infile

    @property
    def _input_source_args(self) -> Iterable[str]:
        return ('-i', self.infile)


# vim:sw=4:ts=4:et:



from typing import Union

from platypush.message.event import Event


class GPIOEvent(Event):
    """
    Event triggered when the value on a GPIO PIN changes.
    """

    def __init__(self, pin: Union[int, str], value: int, *args, **kwargs):
        """
        :param pin: PIN number or name.
        :param value: Current value of the PIN.
        """
        super().__init__(*args, pin=pin, value=value, **kwargs)




import pytest

from platypush.event.hook import EventCondition
from platypush.message.event.assistant import SpeechRecognizedEvent
from platypush.message.event.ping import PingEvent


def test_event_parse():
    """
    Test for the events/conditions matching logic.
    """
    condition = EventCondition.build(
        {
            'type': 'platypush.message.event.ping.PingEvent',
            'message': 'This is a test message',
        }
    )

    event = PingEvent(message=condition.args['message'])
    result = event.matches_condition(condition)
    assert result.is_match

    event = PingEvent(message="This is not a test message")
    result = event.matches_condition(condition)
    assert not result.is_match


def test_nested_event_condition():
    """
    Verify that nested event conditions work as expected.
    """
    condition = EventCondition.build(
        {
            'type': 'platypush.message.event.ping.PingEvent',
            'message': {
                'foo': 'bar',
            },
        }
    )

    event = PingEvent(
        message={
            'foo': 'bar',
            'baz': 'clang',
        }
    )

    assert event.matches_condition(condition).is_match

    event = PingEvent(
        message={
            'something': 'else',
        }
    )

    assert not event.matches_condition(condition).is_match

    event = PingEvent(
        message={
            'foo': 'baz',
        }
    )

    assert not event.matches_condition(condition).is_match


def test_speech_recognized_event_parse():
    """
    Test the event parsing and text extraction logic for the
    SpeechRecognizedEvent.
    """
    condition = EventCondition.build(
        {
            'type': 'platypush.message.event.assistant.SpeechRecognizedEvent',
            'phrase': 'This is (the)? answer: ${answer}',
        }
    )

    event = SpeechRecognizedEvent(phrase="GARBAGE GARBAGE this is the answer: 42")
    result = event.matches_condition(condition)
    assert result.is_match
    assert 'answer' in result.parsed_args
    assert result.parsed_args['answer'] == '42'

    event = SpeechRecognizedEvent(phrase="what is not the answer? 43")
    result = event.matches_condition(condition)
    assert not result.is_match


def test_condition_with_relational_operators():
    """
    Test relational operators used in event conditions.
    """
    # Given: A condition with a relational operator.
    condition = EventCondition.build(
        {
            'type': 'platypush.message.event.ping.PingEvent',
            'message': {'foo': {'$gt': 25}},
        }
    )

    # When: An event with a value greater than 25 is received.
    event = PingEvent(message={'foo': 26})

    # Then: The condition is matched.
    assert event.matches_condition(condition).is_match

    # When: An event with a value lower than 25 is received.
    event = PingEvent(message={'foo': 24})

    # Then: The condition is not matched.
    assert not event.matches_condition(condition).is_match


def test_filter_with_regex_condition():
    """
    Test an event matcher with a regex filter on an attribute.
    """
    # Given: A condition with a regex filter.
    condition = EventCondition.build(
        {
            'type': 'platypush.message.event.ping.PingEvent',
            'message': {'foo': {'$regex': '^ba[rz]'}},
        }
    )

    # When: An event with a matching string is received.
    event = PingEvent(message={'foo': 'bart'})

    # Then: The condition is matched.
    assert event.matches_condition(condition).is_match

    # When: An event with a non-matching string is received.
    event = PingEvent(message={'foo': 'back'})

    # Then: The condition is not matched.
    assert not event.matches_condition(condition).is_match


if __name__ == '__main__':
    pytest.main()


# vim:sw=4:ts=4:et:



import json
from typing import Optional

from tornado.web import stream_request_body

from platypush.backend.http.app.streaming import StreamingRoute

from ._constants import STREAMING_BLOCK_SIZE, STREAMING_CHUNK_SIZE
from ._register import register_media
from ._registry import load_media_map
from ._unregister import unregister_media


@stream_request_body
class MediaStreamRoute(StreamingRoute):
    """
    Route for media streams.
    """

    SUPPORTED_METHODS = ['GET', 'PUT', 'DELETE']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._body = b''

    @classmethod
    def path(cls) -> str:
        return r"^/media/?([a-zA-Z0-9_.]+)?$"

    @property
    def auth_required(self) -> bool:
        return False

    def get(self, media_id: Optional[str] = None):
        """
        Streams a media resource by ID.
        """

        # Return the list of registered streaming media resources if no ID is
        # specified
        if not media_id:
            self.get_media()
            return

        # Strip the extension
        media_id = '.'.join(media_id.split('.')[:-1])

        try:
            self.stream_media(media_id)
        except Exception as e:
            self._on_error(e)

    def put(self, *_, **__):
        """
        The `PUT` route is used to prepare a new media resource for streaming.
        """
        try:
            self.add_media()
        except Exception as e:
            self._on_error(e)

    def delete(self, media_id: Optional[str] = None):
        """
        Removes the given media_id from the map of streaming media.
        """
        media_info = unregister_media(media_id)
        self.write(json.dumps(media_info))

    def data_received(self, chunk: bytes):
        self._body += chunk

    def add_media(self):
        """
        Adds a new media resource to the map of streaming media.
        """
        args = {}
        try:
            args = json.loads(self._body)
        except Exception as e:
            raise AssertionError(f'Invalid JSON request: {e}') from e

        source = args.get('source')
        assert source, 'The request does not contain any source'
        subtitles = args.get('subtitles')
        media_hndl = register_media(source, subtitles)
        ret = media_hndl.to_json()
        if media_hndl.subtitles:
            ret['subtitles_url'] = f'/media/subtitles/{media_hndl.media_id}.vtt'

        self.write(json.dumps(ret))

    def get_media(self):
        """
        Returns the list of registered media resources.
        """
        self.add_header('Content-Type', 'application/json')
        self.finish(json.dumps([dict(media) for media in load_media_map().values()]))

    def stream_media(self, media_id: str):
        """
        Route to stream a media file given its ID.
        """
        media_hndl = load_media_map().get(media_id)
        if not media_hndl:
            raise FileNotFoundError(f'{media_id} is not a registered media_id')

        range_hdr = self.request.headers.get('Range')
        content_length = media_hndl.content_length

        self.add_header('Accept-Ranges', 'bytes')
        self.add_header('Content-Type', media_hndl.mime_type)

        if 'download' in self.request.arguments:
            self.add_header(
                'Content-Disposition',
                'attachment'
                + ('; filename="{media_hndl.filename}"' if media_hndl.filename else ''),
            )

        if range_hdr:
            from_bytes, to_bytes = range_hdr.replace('bytes=', '').split('-')
            from_bytes = int(from_bytes)

            if not to_bytes:
                to_bytes = content_length - 1
                content_length -= from_bytes
            else:
                to_bytes = int(to_bytes)
                content_length = to_bytes - from_bytes

            self.set_status(206)
            self.add_header(
                'Content-Range',
                f'bytes {from_bytes}-{to_bytes}/{media_hndl.content_length}',
            )
        else:
            from_bytes = 0
            to_bytes = STREAMING_BLOCK_SIZE

        self.add_header('Content-Length', str(content_length))
        for chunk in media_hndl.get_data(
            from_bytes=from_bytes,
            to_bytes=to_bytes,
            chunk_size=STREAMING_CHUNK_SIZE,
        ):
            self.write(chunk)
            self.flush()

        self.finish()



from dataclasses import dataclass
import json
import os
import time
from enum import Enum
from threading import Event, RLock
from typing import Optional, Type

import requests

from platypush.config import Config
from platypush.message.event.pushbullet import (
    PushbulletDismissalEvent,
    PushbulletEvent,
    PushbulletFileEvent,
    PushbulletLinkEvent,
    PushbulletMessageEvent,
    PushbulletNotificationEvent,
)
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.pushbullet import PushbulletDeviceSchema, PushbulletSchema


class PushbulletType(Enum):
    """
    PushBullet event types.
    """

    DISMISSAL = 'dismissal'
    FILE = 'file'
    LINK = 'link'
    MESSAGE = 'message'
    MIRROR = 'mirror'
    NOTE = 'note'


@dataclass
class PushbulletEventType:
    """
    PushBullet event type.
    """

    type: PushbulletType
    event_class: Type[PushbulletEvent]


_push_event_types = [
    PushbulletEventType(
        type=PushbulletType.DISMISSAL,
        event_class=PushbulletDismissalEvent,
    ),
    PushbulletEventType(
        type=PushbulletType.FILE,
        event_class=PushbulletFileEvent,
    ),
    PushbulletEventType(
        type=PushbulletType.LINK,
        event_class=PushbulletLinkEvent,
    ),
    PushbulletEventType(
        type=PushbulletType.MESSAGE,
        event_class=PushbulletMessageEvent,
    ),
    PushbulletEventType(
        type=PushbulletType.MIRROR,
        event_class=PushbulletNotificationEvent,
    ),
    PushbulletEventType(
        type=PushbulletType.NOTE,
        event_class=PushbulletMessageEvent,
    ),
]

_push_events_by_type = {t.type.value: t for t in _push_event_types}


class PushbulletPlugin(RunnablePlugin):
    """
    `PushBullet <https://www.pushbullet.com/>`_ integration.

    Among the other things, this plugin allows you to easily interact with your
    mobile devices that have the app installed from Platypush.

    If notification mirroring is enabled on your device, then the push
    notifications will be mirrored to Platypush as well as PushBullet events.

    Since PushBullet also comes with a Tasker integration, you can also use this
    plugin to send commands to your Android device and trigger actions on it.
    It can be used to programmatically send files to your devices and manage
    shared clipboards too.
    """

    _timeout = 15.0
    _upload_timeout = 600.0

    def __init__(
        self,
        token: str,
        device: Optional[str] = None,
        enable_mirroring: bool = True,
        **kwargs,
    ):
        """
        :param token: PushBullet API token, see https://docs.pushbullet.com/#authentication.
        :param device: Device ID that should be exposed. Default: ``Platypush @
            <device_id | hostname>``.
        :param enable_mirroring: If set to True (default) then the plugin will
            receive notifications mirrored from other connected devices -
            these will also be rendered on the connected web clients. Disable
            it if you don't want to forward your mobile notifications through
            the plugin.
        """
        super().__init__(**kwargs)

        if not device:
            device = f'Platypush @ {Config.get_device_id()}'

        self.token = token
        self.device_name = device
        self.enable_mirroring = enable_mirroring
        self.listener = None
        self._initialized = Event()
        self._device = None
        self._init_lock = RLock()
        self._pb = None
        self._device_id = None
        self._devices = []
        self._devices_by_id = {}
        self._devices_by_name = {}

    def _initialize(self):
        from pushbullet import Pushbullet

        if self._initialized.is_set():
            return

        self._pb = Pushbullet(self.token)

        try:
            self._device = self._pb.get_device(self.device_name)
        except Exception as e:
            self.logger.info(
                'Device %s does not exist: %s. Creating it',
                self.device_name,
                e,
            )
            self._device = self._pb.new_device(self.device_name)

        self._device_id = self.get_device_id()
        self._initialized.set()

    @property
    def pb(self):
        """
        :return: PushBullet API object.
        """
        with self._init_lock:
            self._initialize()

        assert self._pb
        return self._pb

    @property
    def device(self):
        """
        :return: Current PushBullet device object.
        """
        with self._init_lock:
            self._initialize()

        assert self._device
        return self._device

    @property
    def device_id(self):
        return self.device.device_iden

    def _request(self, method: str, url: str, **kwargs):
        meth = getattr(requests, method)
        resp = meth(
            'https://api.pushbullet.com/v2/' + url.lstrip('/'),
            timeout=self._timeout,
            headers={
                'Authorization': 'Bearer ' + self.token,
                'Content-Type': 'application/json',
            },
            **kwargs,
        )

        resp.raise_for_status()
        return resp.json()

    def get_device_id(self):
        assert self._pb

        try:
            return self._pb.get_device(self.device_name).device_iden
        except Exception:
            device = self.pb.new_device(
                self.device_name,
                model='Platypush virtual device',
                manufacturer='platypush',
                icon='system',
            )

            self.logger.info('Created Pushbullet device %s', self.device_name)
            return device.device_iden

    def _get_latest_push(self):
        t = int(time.time()) - 10
        pushes = self.pb.get_pushes(modified_after=str(t), limit=1)
        if pushes:
            return pushes[0]

        return None

    def on_open(self, *_, **__):
        self.logger.info('Pushbullet connected')

    def on_close(self, args=None):
        err = args[0] if args else None
        self.close()
        assert not err or self.should_stop(), 'Pushbullet connection closed: ' + str(
            err or 'unknown error'
        )

    def on_error(self, *args):
        raise RuntimeError('Pushbullet error: ' + str(args))

    def on_push(self, data):
        try:
            # Parse the push
            try:
                data = json.loads(data) if isinstance(data, str) else data
            except Exception as e:
                self.logger.exception(e)
                return

            # If it's a push, get it
            push = None
            if data['type'] == 'tickle' and data['subtype'] == 'push':
                push = self._get_latest_push()
            elif data['type'] == 'push':
                push = data['push']

            if not push:
                self.logger.debug('Not a push notification.\nMessage: %s', data)
                return

            push_type = push.pop('type', None)
            push_event_type = _push_events_by_type.get(push_type)
            if not push_event_type:
                self.logger.debug(
                    'Unknown push type: %s.\nMessage: %s', push_type, data
                )
                return

            if (
                not self.enable_mirroring
                and push_event_type.type == PushbulletType.MIRROR
            ):
                return

            push = dict(PushbulletSchema().dump(push))
            evt_type = push_event_type.event_class
            self._bus.post(evt_type(**push))
        except Exception as e:
            self.logger.warning(
                'Error while processing push: %s.\nMessage: %s', e, data
            )
            self.logger.exception(e)
            return

    def _get_device(self, device) -> Optional[dict]:
        output = None
        refreshed = False

        while not output:
            if device in self._devices_by_id:
                return self._devices_by_id[device]
            if device in self._devices_by_name:
                return self._devices_by_name[device]
            if refreshed:
                return None

            self.get_devices()
            refreshed = True

        return None

    def close(self):
        if self.listener:
            try:
                self.listener.close()
            except Exception:
                pass

            self.listener = None

        if self._pb:
            self._pb = None

        self._initialized.clear()

    def run_listener(self):
        from .listener import Listener

        self.listener = Listener(
            account=self.pb,
            on_push=self.on_push,
            on_open=self.on_open,
            on_close=self.on_close,
            on_error=self.on_error,
        )

        self.listener.run_forever()

    @action
    def get_devices(self):
        """
        Get the list of available devices.

        :return: .. schema:: pushbullet.PushbulletDeviceSchema(many=True)
        """
        resp = self._request('get', 'devices')
        self._devices = resp.get('devices', [])
        self._devices_by_id = {dev['iden']: dev for dev in self._devices}
        self._devices_by_name = {
            dev['nickname']: dev for dev in self._devices if 'nickname' in dev
        }

        return PushbulletDeviceSchema(many=True).dump(self._devices)

    @action
    def get_device(self, device: str) -> Optional[dict]:
        """
        Get a device by ID or name.

        :param device: Device ID or name
        :return: .. schema:: pushbullet.PushbulletDeviceSchema
        """
        dev = self._get_device(device)
        if not dev:
            return None

        return dict(PushbulletDeviceSchema().dump(dev))

    @action
    def get_pushes(self, limit: int = 10):
        """
        Get the list of pushes.

        :param limit: Maximum number of pushes to fetch (default: 10).
        :return: .. schema:: pushbullet.PushbulletSchema(many=True)
        """
        return PushbulletSchema().dump(
            self._request('get', 'pushes', params={'limit': limit}).get('pushes', []),
            many=True,
        )

    @action
    def send_note(
        self,
        device: Optional[str] = None,
        body: Optional[str] = None,
        title: Optional[str] = None,
        url: Optional[str] = None,
        **kwargs,
    ):
        """
        Send a note push.

        :param device: Device ID or name (default: None, all devices)
        :param body: Note body
        :param title: Note title
        :param url: URL attached to the note
        :param kwargs: Push arguments, see https://docs.pushbullet.com/#create-push
        :return: .. schema:: pushbullet.PushbulletSchema
        """

        dev = None
        if device:
            dev = self._get_device(device)
            assert dev, f'No such device: {device}'

        kwargs['body'] = body
        kwargs['title'] = title
        kwargs['type'] = 'note'
        if url:
            kwargs['type'] = 'link'
            kwargs['url'] = url

        if dev:
            kwargs['device_iden'] = dev['iden']

        rs = self._request('post', 'pushes', data=json.dumps(kwargs))
        return dict(PushbulletSchema().dump(rs))

    @action
    def send_file(self, filename: str, device: Optional[str] = None):
        """
        Send a file.

        :param device: Device ID or name (default: None, all devices)
        :param filename: Path to the local file
        """

        dev = None
        if device:
            dev = self._get_device(device)
            assert dev, f'No such device: {device}'

        upload_req = self._request(
            'post',
            'upload-request',
            data=json.dumps({'file_name': os.path.basename(filename)}),
        )

        with open(filename, 'rb') as f:
            rs = requests.post(
                upload_req['upload_url'],
                data=upload_req['data'],
                files={'file': f},
                timeout=self._upload_timeout,
            )

        rs.raise_for_status()
        self._request(
            'post',
            'pushes',
            data=json.dumps(
                {
                    'type': 'file',
                    'device_iden': dev['iden'] if dev else None,
                    'file_name': upload_req['file_name'],
                    'file_type': upload_req.get('file_type'),
                    'file_url': upload_req['file_url'],
                }
            ),
        )

        return {
            'filename': upload_req['file_name'],
            'type': upload_req.get('file_type'),
            'url': upload_req['file_url'],
        }

    @action
    def send_clipboard(self, text: str):
        """
        Send text to the clipboard of other devices.

        :param text: Text to be copied.
        """
        self._request(
            'post',
            'ephemerals',
            data=json.dumps(
                {
                    'type': 'push',
                    'push': {
                        'body': text,
                        'type': 'clip',
                        'source_device_iden': self.device_id,
                    },
                }
            ),
        )

    def main(self):
        while not self.should_stop():
            while not self._initialized.is_set():
                try:
                    self._initialize()
                except Exception as e:
                    self.logger.exception(e)
                    self.logger.error('Pushbullet initialization error: %s', e)
                    self.wait_stop(10)

            while not self.should_stop():
                try:
                    self.run_listener()
                except Exception as e:
                    if not self.should_stop():
                        self.logger.exception(e)
                        self.logger.error('Pushbullet listener error: %s', e)

                    self.wait_stop(10)


# vim:sw=4:ts=4:et:



from sqlalchemy import Column, Integer, ForeignKey, Boolean, String, JSON

from platypush.common.db import is_defined

from .devices import Device


if not is_defined('switch'):

    class Switch(Device):
        __tablename__ = 'switch'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        state = Column(Boolean)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('enum_switch'):

    class EnumSwitch(Device):
        __tablename__ = 'enum_switch'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        value = Column(String)
        values = Column(JSON)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



import os
import requests
from typing import Optional, Union, Iterable, Mapping, Sequence

from platypush.plugins import Plugin, action
from platypush.schemas.mastodon import MastodonSchema, MastodonSearchSchema, MastodonAccountCreationSchema, \
    MastodonAccountSchema, MastodonStatusSchema, MastodonFeaturedHashtagSchema, MastodonAccountListSchema, \
    MastodonFilterSchema, MastodonMediaSchema, MastodonConversationSchema, MastodonListSchema, \
    MastodonNotificationSchema
from platypush.utils import get_mime_type


class MastodonPlugin(Plugin):
    """
    Plugin to interact with `Mastodon <https://mastodon.social/about>`_ instances.

    It requires an active API token associated to an app registered on the instance.
    In order to get one:

        - Open ``https://<mastodon-base-url>/settings/applications/``
        - Create a new application
        - Select the scopes relevant for your specific usage.
        - Take note of the token reported on the *Your access token* row.

    The notifications subscription service requires the ``ngrok`` plugin and the
    `http` backend to be enabled, since we need to expose an external URL that
    the Mastodon instance can call when new events occur.
    """

    class SubscriptionConfig:
        tunnel_url: str
        local_port: int
        auth_secret: str
        private_key: str
        public_key: str
        server_key: str

    def __init__(self, base_url: str, access_token: Optional[str] = None, **kwargs):
        """
        :param base_url: Base URL of the Mastodon web server, in the form of ``https://<domain-name>``.
        :param access_token: Access token as reported on ``https://<base_url>/settings/applications/<app_id>``.
        """
        super().__init__(**kwargs)
        self._base_url = base_url
        self._access_token = access_token
        self._subscription_config = self.SubscriptionConfig()

    def base_url(self, version: str, base_url: Optional[str] = None) -> str:
        return f'{base_url or self._base_url}/api/{version}'

    def _run(
            self, path: str, method: str = 'get', version: str = 'v2', headers: Optional[dict] = None,
            base_url: Optional[str] = None, access_token: Optional[str] = None,
            schema: Optional[MastodonSchema] = None, **kwargs
    ) -> Optional[Union[dict, list]]:
        headers = {
            'Authorization': f'Bearer {access_token or self._access_token}',
            'Accept': 'application/json',
            **(headers or {}),
        }

        method = getattr(requests, method.lower())
        rs = method(self.base_url(base_url=base_url, version=version) + '/' + path, headers=headers, **kwargs)
        rs.raise_for_status()
        rs = rs.json()
        if schema:
            rs = schema.dump(rs)
        return rs

    # noinspection PyShadowingBuiltins
    @action
    def search(
            self, query: str, type: Optional[str] = None, min_id: Optional[str] = None,
            max_id: Optional[str] = None, limit: int = 20, offset: int = 0, following: bool = False,
            **kwargs
    ) -> Mapping[str, Iterable[dict]]:
        """
        Perform a search.

        :param query: Search query.
        :param type: Filter by type. Supported types:

            - ``accounts``
            - ``hashtags``
            - ``statuses``

        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param following: Only return results from accounts followed by the user (default: False).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonSearchSchema
        """
        return self._run(
            'search',
            version='v2',
            schema=MastodonSearchSchema(),
            params={
                'q': query,
                **({'type': type} if type else {}),
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
                **({'following': following} if following else {}),
            }, **kwargs
        )

    @action
    def register_account(
            self, username: str, email: str, password: str, locale: str = 'en',
            reason: Optional[str] = None, **kwargs
    ) -> dict:
        """
        Register a new account.

        It requires the specified API token to have ``write:accounts`` permissions.

        :param username: User name.
        :param email: User's email address (must be a valid address).
        :param password: The password used for the first login.
        :param locale: Language/encoding for the confirmation email.
        :param reason: Text that will be reviewed by moderators if registrations require manual approval.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountCreationSchema
        """
        return self._run(
            'accounts',
            method='post',
            version='v1',
            schema=MastodonAccountCreationSchema(),
            json={
                'username': username,
                'email': email,
                'password': password,
                'locale': locale,
                'reason': reason,
                'agreement': True,
            }, **kwargs
        )

    @action
    def update_account(
            self, discoverable: Optional[bool] = None, bot: Optional[bool] = None,
            display_name: Optional[str] = None, note: Optional[str] = None,
            avatar: Optional[str] = None, header: Optional[str] = None,
            locked: Optional[bool] = None, privacy: Optional[str] = None,
            sensitive: Optional[bool] = None, language: Optional[str] = None,
            metadata: Optional[Iterable[Mapping]] = None, **kwargs
    ) -> dict:
        """
        Updates the properties of the account associated to the access token.

        It requires the specified API token to have ``write:accounts`` permissions.

        :param discoverable: Whether the account should be shown in the profile directory.
        :param bot: Whether the account is a bot.
        :param display_name: The display name to use for the profile.
        :param note: The account bio (HTML is supported).
        :param avatar: Path to an avatar image.
        :param header: Path to a header image.
        :param locked: Whether manual approval of follow requests is required.
        :param privacy: Default post privacy for authored statuses.
        :param sensitive: Whether to mark authored statuses as sensitive by default.
        :param language: Default language to use for authored statuses (ISO 6391 code).
        :param metadata: Profile metadata items with ``name`` and ``value``.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountSchema
        """
        avatar = os.path.expanduser(avatar) if avatar else None
        header = os.path.expanduser(header) if header else None
        return self._run(
            'accounts/update_credentials',
            method='patch',
            version='v1',
            schema=MastodonAccountSchema(),
            data={
                **({'discoverable': discoverable} if discoverable is not None else {}),
                **({'bot': bot} if bot is not None else {}),
                **({'display_name': display_name} if display_name is not None else {}),
                **({'note': note} if note is not None else {}),
                **({'locked': locked} if locked is not None else {}),
                **({'source[privacy]': privacy} if privacy is not None else {}),
                **({'source[sensitive]': sensitive} if sensitive is not None else {}),
                **({'source[language]': language} if language is not None else {}),
                **({'fields_attributes': metadata} if metadata is not None else {}),
            },
            files={
                **({'avatar': (
                    os.path.basename(avatar), open(avatar, 'rb'), get_mime_type(avatar)
                )} if avatar is not None else {}),
                **({'header': (
                    os.path.basename(header), open(header, 'rb'), get_mime_type(header)
                )} if header is not None else {}),
            },
            **kwargs
        )

    @action
    def get_account(self, account_id: str, **kwargs) -> dict:
        """
        Retrieve an account by ID.

        It requires the specified API token to have ``read:accounts`` permissions.

        :param account_id: Account ID to retrieve.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountSchema
        """
        return self._run(
            f'accounts/{account_id}',
            version='v1',
            schema=MastodonAccountSchema(),
            **kwargs
        )

    @action
    def get_statuses(self, account_id: str, min_id: Optional[str] = None, max_id: Optional[str] = None,
                     limit: int = 20, offset: int = 0, **kwargs) -> Iterable[dict]:
        """
        Retrieve statuses by account ID.

        It requires the specified API token to have the ``read:statuses`` permission.

        :param account_id: Account ID.
        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema(many=True)
        """
        return self._run(
            f'accounts/{account_id}/statuses',
            version='v1',
            schema=MastodonStatusSchema(many=True),
            params={
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            },
            **kwargs
        )

    @action
    def get_followers(self, account_id: str, max_id: Optional[str] = None,
                      since_id: Optional[str] = None, limit: int = 20, offset: int = 0,
                      **kwargs) -> Iterable[dict]:
        """
        Retrieve the list of followers of an account.

        It requires the specified API token to have the ``read:accounts`` permission.

        :param account_id: Account ID.
        :param max_id: Return results older than this ID.
        :param since_id: Return results newer than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountSchema(many=True)
        """
        return self._run(
            f'accounts/{account_id}/followers',
            version='v1',
            schema=MastodonAccountSchema(many=True),
            params={
                **({'since_id': since_id} if since_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            },
            **kwargs
        )

    @action
    def get_following(self, account_id: str, max_id: Optional[str] = None,
                      since_id: Optional[str] = None, limit: int = 20, offset: int = 0,
                      **kwargs) -> Iterable[dict]:
        """
        Retrieve the list of accounts followed by a specified account.

        It requires the specified API token to have the ``read:accounts`` permission.

        :param account_id: Account ID.
        :param max_id: Return results older than this ID.
        :param since_id: Return results newer than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountSchema(many=True)
        """
        return self._run(
            f'accounts/{account_id}/following',
            version='v1',
            schema=MastodonAccountSchema(many=True),
            params={
                **({'since_id': since_id} if since_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            },
            **kwargs
        )

    @action
    def get_featured_tags(self, account_id: Optional[str] = None, max_id: Optional[str] = None,
                          since_id: Optional[str] = None, limit: int = 20, offset: int = 0,
                          **kwargs) -> Iterable[dict]:
        """
        Retrieve the list of featured hashtags of an account.

        It requires the specified API token to have the ``read:accounts`` permission.

        :param account_id: Account ID (if not specified then retrieve the featured tags of the current account).
        :param max_id: Return results older than this ID.
        :param since_id: Return results newer than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonFeaturedHashtagSchema(many=True)
        """
        return self._run(
            f'accounts/{account_id}/featured_tags' if account_id else 'featured_tags',
            version='v1',
            schema=MastodonFeaturedHashtagSchema(many=True),
            params={
                **({'since_id': since_id} if since_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            },
            **kwargs
        )

    @action
    def get_featured_lists(self, account_id: str, max_id: Optional[str] = None,
                           since_id: Optional[str] = None, limit: int = 20, offset: int = 0,
                           **kwargs) -> Iterable[dict]:
        """
        Retrieve the list that you have added a certain account to.

        It requires the specified API token to have the ``read:lists`` permission.

        :param account_id: Account ID.
        :param max_id: Return results older than this ID.
        :param since_id: Return results newer than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountListSchema(many=True)
        """
        return self._run(
            f'accounts/{account_id}/lists',
            version='v1',
            schema=MastodonAccountListSchema(many=True),
            params={
                **({'since_id': since_id} if since_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            },
            **kwargs
        )

    @action
    def follow_account(self, account_id: str, notify: bool = False, reblogs: bool = True, **kwargs):
        """
        Follow a given account ID.

        It requires the specified API token to have the ``write:follows`` permission.

        :param account_id: Account ID.
        :param notify: Receive notifications when this account posts a new status (default: False).
        :param reblogs: Receive this account's reblogs on your timeline (default: True).
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/follow',
            version='v1',
            method='post',
            json={'notify': notify, 'reblogs': reblogs},
            **kwargs
        )

    @action
    def unfollow_account(self, account_id: str, **kwargs):
        """
        Unfollow a given account ID.

        It requires the specified API token to have the ``write:follows`` permission.

        :param account_id: Account ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/unfollow',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def block_account(self, account_id: str, **kwargs):
        """
        Block a given account ID.

        It requires the specified API token to have the ``write:blocks`` permission.

        :param account_id: Account ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/block',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def unblock_account(self, account_id: str, **kwargs):
        """
        Unblock a given account ID.

        It requires the specified API token to have the ``write:blocks`` permission.

        :param account_id: Account ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/unblock',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def mute_account(self, account_id: str, **kwargs):
        """
        Mute a given account ID.

        It requires the specified API token to have the ``write:mutes`` permission.

        :param account_id: Account ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/mute',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def unmute_account(self, account_id: str, **kwargs):
        """
        Unmute a given account ID.

        It requires the specified API token to have the ``write:mutes`` permission.

        :param account_id: Account ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/unmute',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def pin_account(self, account_id: str, **kwargs):
        """
        Pin a given account ID to your profile.

        It requires the specified API token to have the ``write:accounts`` permission.

        :param account_id: Account ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/pin',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def unpin_account(self, account_id: str, **kwargs):
        """
        Unpin a given account ID from your profile.

        It requires the specified API token to have the ``write:accounts`` permission.

        :param account_id: Account ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/unpin',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def set_account_note(self, account_id: str, note: str, **kwargs):
        """
        Set a private note for an account.

        It requires the specified API token to have the ``write:accounts`` permission.

        :param account_id: Account ID.
        :param note: Note content (HTML is supported).
        :param kwargs: ``base_url``/``access_token`` override.
        """
        self._run(
            f'accounts/{account_id}/note',
            version='v1',
            method='post',
            json={'comment': note},
            **kwargs
        )

    @action
    def get_bookmarked_statuses(self, min_id: Optional[str] = None,
                                max_id: Optional[str] = None, limit: int = 20, **kwargs) -> Iterable[dict]:
        """
        Retrieve the list of statuses bookmarked by the user.

        It requires the specified API token to have the ``read:bookmarks`` permission.

        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema(many=True)
        """
        return self._run(
            'bookmarks',
            version='v1',
            schema=MastodonStatusSchema(many=True),
            params={
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
            },
            **kwargs
        )

    @action
    def get_favourited_statuses(self, min_id: Optional[str] = None,
                                max_id: Optional[str] = None, limit: int = 20, **kwargs) -> Iterable[dict]:
        """
        Retrieve the list of statuses favourited by the account.

        It requires the specified API token to have the ``read:favourites`` permission.

        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema(many=True)
        """
        return self._run(
            'favourites',
            version='v1',
            schema=MastodonStatusSchema(many=True),
            params={
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
            },
            **kwargs
        )

    @action
    def get_muted_accounts(self, max_id: Optional[str] = None,
                           since_id: Optional[str] = None, limit: int = 20,
                           **kwargs) -> Iterable[dict]:
        """
        Retrieve the list of muted accounts.

        It requires the specified API token to have the ``read:mutes`` permission.

        :param max_id: Return results older than this ID.
        :param since_id: Return results newer than this ID.
        :param limit: Maximum number of results (default: 20).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountSchema(many=True)
        """
        return self._run(
            'mutes',
            version='v1',
            schema=MastodonAccountSchema(many=True),
            params={
                **({'since_id': since_id} if since_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
            },
            **kwargs
        )

    @action
    def get_blocked_accounts(self, max_id: Optional[str] = None,
                             since_id: Optional[str] = None, limit: int = 20,
                             **kwargs) -> Iterable[dict]:
        """
        Retrieve the list of blocked accounts.

        It requires the specified API token to have the ``read:blocks`` permission.

        :param max_id: Return results older than this ID.
        :param since_id: Return results newer than this ID.
        :param limit: Maximum number of results (default: 20).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountSchema(many=True)
        """
        return self._run(
            'blocks',
            version='v1',
            schema=MastodonAccountSchema(many=True),
            params={
                **({'since_id': since_id} if since_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
            },
            **kwargs
        )

    @action
    def get_filters(self, **kwargs) -> Iterable[dict]:
        """
        Retrieve the list of filters created by the account.

        It requires the specified API token to have the ``read:filters`` permission.

        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonFilterSchema(many=True)
        """
        return self._run(
            'filters',
            version='v1',
            schema=MastodonFilterSchema(many=True),
            **kwargs
        )

    @action
    def create_filter(self, phrase: str, context: Iterable[str],
                      irreversible: Optional[bool] = None,
                      whole_word: Optional[bool] = None,
                      expires_in: Optional[int] = None,
                      **kwargs) -> dict:
        """
        Create a new filter.

        It requires the specified API token to have the ``write:filters`` permission.

        :param phrase: Text to be filtered.
        :param context: Array of enumerable strings: ``home``, ``notifications``, ``public``, ``thread``.
            At least one context must be specified.
        :param irreversible: Should the server irreversibly drop matching entities from home and notifications?
        :param whole_word: Consider word boundaries?
        :param expires_in: Expires in the specified number of seconds.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonFilterSchema
        """
        return self._run(
            'filters',
            version='v1',
            method='post',
            schema=MastodonFilterSchema(),
            json={
                'phrase': phrase,
                'context': context,
                **({'irreversible': irreversible} if irreversible is not None else {}),
                **({'whole_word': whole_word} if whole_word is not None else {}),
                **({'expires_in': expires_in} if expires_in is not None else {}),
            },
            **kwargs
        )

    @action
    def update_filter(self, filter_id: int,
                      phrase: Optional[str] = None,
                      context: Optional[Iterable[str]] = None,
                      irreversible: Optional[bool] = None,
                      whole_word: Optional[bool] = None,
                      expires_in: Optional[int] = None,
                      **kwargs) -> dict:
        """
        Update a filter.

        It requires the specified API token to have the ``write:filters`` permission.

        :param filter_id: Filter ID.
        :param phrase: Text to be filtered.
        :param context: Array of enumerable strings: ``home``, ``notifications``, ``public``, ``thread``.
            At least one context must be specified.
        :param irreversible: Should the server irreversibly drop matching entities from home and notifications?
        :param whole_word: Consider word boundaries?
        :param expires_in: Expires in the specified number of seconds.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonFilterSchema
        """
        return self._run(
            f'filters/{filter_id}',
            version='v1',
            method='put',
            schema=MastodonFilterSchema(),
            json={
                **({'phrase': phrase} if phrase is not None else {}),
                **({'context': context} if context is not None else {}),
                **({'irreversible': irreversible} if irreversible is not None else {}),
                **({'whole_word': whole_word} if whole_word is not None else {}),
                **({'expires_in': expires_in} if expires_in is not None else {}),
            },
            **kwargs
        )

    @action
    def remove_filter(self, filter_id: int, **kwargs):
        """
        Remove a filter.

        It requires the specified API token to have the ``write:filters`` permission.

        :param filter_id: Filter ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'filters/{filter_id}',
            version='v1',
            method='delete',
            **kwargs
        )

    @action
    def add_featured_tag(self, name: str, **kwargs) -> dict:
        """
        Add a featured tag to the current account.

        It requires the specified API token to have the ``write:accounts`` permission.

        :param name: Hashtag name.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonFeaturedHashtagSchema
        """
        return self._run(
            'featured_tags',
            version='v1',
            method='post',
            schema=MastodonFeaturedHashtagSchema(),
            json={'name': name},
            **kwargs
        )

    @action
    def remove_featured_tag(self, tag_id: int, **kwargs):
        """
        Remove a featured tag from the current account.

        It requires the specified API token to have the ``write:accounts`` permission.

        :param tag_id: Hashtag ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'featured_tags/{tag_id}',
            version='v1',
            method='delete',
            **kwargs
        )

    @action
    def publish_status(self, status: str, in_reply_to_id: Optional[str] = None,
                       media_ids: Optional[Iterable[str]] = None,
                       sensitive: Optional[bool] = None, spoiler_text: Optional[str] = None,
                       visibility: Optional[str] = None, **kwargs) -> dict:
        """
        Publish a new status.

        It requires the specified API token to have the ``write:statuses`` permission.

        :param status: Content of the status to publish.
        :param in_reply_to_id: Post the status in reply to this status ID.
        :param media_ids: Optional list of media IDs to add as attachments.
        :param sensitive: Set to true if sensitive.
        :param spoiler_text: Set for optional spoiler text.
        :param visibility: Supported values: ``public``, ``unlisted``, ``private`` and ``direct``.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema
        """
        return self._run(
            'statuses',
            version='v1',
            method='post',
            schema=MastodonStatusSchema(),
            json={
                'status': status,
                **({'in_reply_to_id': in_reply_to_id} if in_reply_to_id is None else {}),
                **({'media_ids': media_ids} if media_ids else {}),
                **({'sensitive': sensitive} if sensitive is None else {}),
                **({'spoiler_text': spoiler_text} if spoiler_text is None else {}),
                **({'visibility': visibility} if visibility is None else {}),
            },
            **kwargs
        )

    @action
    def get_status(self, status_id: str, **kwargs) -> dict:
        """
        Get a status by ID.

        It requires the specified API token to have the ``read:statuses`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema
        """
        return self._run(
            f'statuses/{status_id}',
            version='v1',
            schema=MastodonStatusSchema(),
            **kwargs
        )

    @action
    def remove_status(self, status_id: str, **kwargs):
        """
        Remove a status by ID.

        It requires the specified API token to have the ``read:statuses`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}',
            version='v1',
            method='delete',
            **kwargs
        )

    @action
    def add_favourite_status(self, status_id: str, **kwargs):
        """
        Favourite a status.

        It requires the specified API token to have the ``write:favourites`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/favourite',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def remove_favourite_status(self, status_id: str, **kwargs):
        """
        Undo a status favourite action.

        It requires the specified API token to have the ``write:favourites`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/favourite',
            version='v1',
            method='delete',
            **kwargs
        )

    @action
    def reblog_status(self, status_id: str, **kwargs):
        """
        Reblog (a.k.a. reshare/boost) a status.

        It requires the specified API token to have the ``write:statuses`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/reblog',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def undo_reblog_status(self, status_id: str, **kwargs):
        """
        Undo a status reblog.

        It requires the specified API token to have the ``write:statuses`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/unreblog',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def bookmark_status(self, status_id: str, **kwargs):
        """
        Add a status to the bookmarks.

        It requires the specified API token to have the ``write:bookmarks`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/bookmark',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def undo_bookmark_status(self, status_id: str, **kwargs):
        """
        Remove a status from the bookmarks.

        It requires the specified API token to have the ``write:bookmarks`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/unbookmark',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def mute_status(self, status_id: str, **kwargs):
        """
        Mute updates on a status.

        It requires the specified API token to have the ``write:mutes`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/mute',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def unmute_status(self, status_id: str, **kwargs):
        """
        Restore updates on a status.

        It requires the specified API token to have the ``write:mutes`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/unmute',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def pin_status(self, status_id: str, **kwargs):
        """
        Pin a status to the profile.

        It requires the specified API token to have the ``write:accounts`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/pin',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def unpin_status(self, status_id: str, **kwargs):
        """
        Remove a pinned status.

        It requires the specified API token to have the ``write:accounts`` permission.

        :param status_id: Status ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'statuses/{status_id}/unpin',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def upload_media(self, file: str, description: Optional[str] = None,
                     thumbnail: Optional[str] = None, **kwargs) -> dict:
        """
        Upload media that can be used as attachments.

        It requires the specified API token to have the ``write:media`` permission.

        :param file: Path to the file to upload.
        :param thumbnail: Path to the file thumbnail.
        :param description: Optional attachment description.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonMediaSchema
        """
        file_path = os.path.expanduser(file)
        thumbnail_path = os.path.expanduser(thumbnail) if thumbnail else None
        return self._run(
            'media',
            version='v1',
            method='post',
            schema=MastodonMediaSchema(),
            data={
                **({'description': description} if description else {}),
            },
            files={
                'file': (
                    os.path.basename(file_path), open(file_path, 'rb'), get_mime_type(file_path)
                ),
                **(
                    {
                        'thumbnail': (
                            os.path.basename(thumbnail_path),
                            open(os.path.expanduser(thumbnail_path), 'rb'),
                            get_mime_type(thumbnail_path)
                        ),
                    } if thumbnail_path else {}
                ),
            },
            **kwargs
        )

    @action
    def update_media(self, media_id: str, file: Optional[str] = None, description: Optional[str] = None,
                     thumbnail: Optional[str] = None, **kwargs) -> dict:
        """
        Update a media attachment.

        It requires the specified API token to have the ``write:media`` permission.

        :param media_id: Media ID to update.
        :param file: Path to the new file.
        :param description: New description.
        :param thumbnail: Path to the new thumbnail.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonMediaSchema
        """
        file = os.path.expanduser(file)
        thumbnail = os.path.expanduser(thumbnail)
        return self._run(
            f'media/{media_id}',
            version='v1',
            method='put',
            schema=MastodonMediaSchema(),
            data={
                **({'description': description} if description else {}),
            },
            files={
                'file': (
                    os.path.basename(file), open(file, 'rb'), get_mime_type(file)
                ),
                **(
                    {
                        'thumbnail': (
                            os.path.basename(thumbnail),
                            open(os.path.expanduser(thumbnail), 'rb'),
                            get_mime_type(thumbnail)
                        ),
                    } if thumbnail else {}
                ),
            },
            **kwargs
        )

    @action
    def get_public_timeline(
            self, local: bool = False, remote: bool = False, only_media: bool = False,
            min_id: Optional[str] = None, max_id: Optional[str] = None, limit: int = 20,
            offset: int = 0, **kwargs
    ) -> Iterable[dict]:
        """
        Get a list of statuses from the public timeline.

        It requires the specified API token to have the ``read:statuses`` permission.

        :param local: Retrieve only local statuses (default: ``False``).
        :param remote: Retrieve only remote statuses (default: ``False``).
        :param only_media:  Retrieve only statuses with media attached (default: ``False``).
        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema(many=True)
        """
        return self._run(
            'timelines/public',
            version='v1',
            schema=MastodonStatusSchema(many=True),
            params={
                **({'local': local} if local else {}),
                **({'remote': remote} if remote else {}),
                **({'only_media': only_media} if only_media else {}),
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            }, **kwargs
        )

    @action
    def get_hashtag_timeline(
            self, hashtag: str, local: bool = False, only_media: bool = False,
            min_id: Optional[str] = None, max_id: Optional[str] = None, limit: int = 20,
            offset: int = 0, **kwargs
    ) -> Iterable[dict]:
        """
        Get a list of statuses associated to a hashtag.

        It requires the specified API token to have the ``read:statuses`` permission.

        :param hashtag: Hashtag to search.
        :param local: Retrieve only local statuses (default: ``False``).
        :param only_media:  Retrieve only statuses with media attached (default: ``False``).
        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema(many=True)
        """
        return self._run(
            f'timelines/tag/{hashtag}',
            version='v1',
            schema=MastodonStatusSchema(many=True),
            params={
                **({'local': local} if local else {}),
                **({'only_media': only_media} if only_media else {}),
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            }, **kwargs
        )

    @action
    def get_home_timeline(
            self, local: bool = False, only_media: bool = False,
            min_id: Optional[str] = None, max_id: Optional[str] = None, limit: int = 20,
            offset: int = 0, **kwargs
    ) -> Iterable[dict]:
        """
        Get a list of statuses from the followed users.

        It requires the specified API token to have the ``read:statuses`` permission.

        :param local: Retrieve only local statuses (default: ``False``).
        :param only_media:  Retrieve only statuses with media attached (default: ``False``).
        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema(many=True)
        """
        return self._run(
            f'timelines/home',
            version='v1',
            schema=MastodonStatusSchema(many=True),
            params={
                **({'local': local} if local else {}),
                **({'only_media': only_media} if only_media else {}),
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            }, **kwargs
        )

    @action
    def get_list_timeline(
            self, list_id: str,
            min_id: Optional[str] = None, max_id: Optional[str] = None, limit: int = 20,
            offset: int = 0, **kwargs
    ) -> Iterable[dict]:
        """
        Get a list of statuses from a list timeline.

        It requires the specified API token to have the ``read:lists`` permission.

        :param list_id: List ID.
        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param offset: Return results from this offset (default: 0).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonStatusSchema(many=True)
        """
        return self._run(
            f'timelines/list/{list_id}',
            version='v1',
            schema=MastodonStatusSchema(many=True),
            params={
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
                **({'offset': offset} if offset else {}),
            }, **kwargs
        )

    @action
    def get_conversations(
            self, min_id: Optional[str] = None, max_id: Optional[str] = None,
            limit: int = 20,  **kwargs
    ) -> Iterable[dict]:
        """
        Get a list of user conversations.

        It requires the specified API token to have the ``read:statuses`` permission.

        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonConversationSchema(many=True)
        """
        return self._run(
            'conversations',
            version='v1',
            schema=MastodonConversationSchema(many=True),
            params={
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
            }, **kwargs
        )

    @action
    def remove_conversation(self, conversation_id: int, **kwargs):
        """
        Remove a conversation by ID.

        It requires the specified API token to have the ``write_conversations`` permission.

        :param conversation_id: Conversation ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'conversations/{conversation_id}',
            version='v1',
            method='delete',
            **kwargs
        )

    @action
    def mark_conversation_as_read(self, conversation_id: int, **kwargs):
        """
        Mark a conversation as read.

        It requires the specified API token to have the ``write_conversations`` permission.

        :param conversation_id: Conversation ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'conversations/{conversation_id}/read',
            version='v1',
            method='post',
            **kwargs
        )

    @action
    def get_lists(self, list_id: Optional[int] = None, **kwargs) -> Union[dict, Iterable[dict]]:
        """
        Get the lists owned by the logged user.

        It requires the specified API token to have the ``read:lists`` permission.

        :param list_id: Retrieve a specific list ID (default: retrieve all).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonListSchema(many=True)
        """
        return self._run(
            'lists' + (f'/{list_id}' if list_id else ''),
            version='v1',
            method='get',
            schema=MastodonListSchema(many=list_id is None),
            **kwargs
        )

    @action
    def create_list(self, title: str, replies_policy: str = 'list', **kwargs) -> dict:
        """
        Create a new list.

        It requires the specified API token to have the ``write:lists`` permission.

        :param title: List title.
        :param replies_policy: Possible values: ``none``, ``following`` or ``list``. Default: ``list``.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonListSchema
        """
        return self._run(
            'lists',
            version='v1',
            method='post',
            schema=MastodonListSchema(),
            data={'title': title, 'replies_policy': replies_policy},
            **kwargs
        )

    @action
    def update_list(
            self, list_id: int, title: Optional[str], replies_policy: Optional[str] = None, **kwargs
    ) -> dict:
        """
        Update a list.

        It requires the specified API token to have the ``write:lists`` permission.

        :param list_id: List ID.
        :param title: New list title.
        :param replies_policy: New replies policy.
            Possible values: ``none``, ``following`` or ``list``. Default: ``list``.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonListSchema
        """
        return self._run(
            f'lists/{list_id}',
            version='v1',
            method='put',
            schema=MastodonListSchema(),
            data={
                **({'title': title} if title else {}),
                **({'replies_policy': replies_policy} if replies_policy else {}),
            },
            **kwargs
        )

    @action
    def delete_list(self, list_id: int, **kwargs):
        """
        Delete a list.

        It requires the specified API token to have the ``write:lists`` permission.

        :param list_id: List ID.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            f'lists/{list_id}',
            version='v1',
            method='delete',
            **kwargs
        )

    @action
    def get_list_accounts(
            self, list_id: Optional[int] = None, min_id: Optional[str] = None,
            max_id: Optional[str] = None, limit: int = 20, **kwargs
    ) -> Iterable[dict]:
        """
        Get the accounts in a list.

        It requires the specified API token to have the ``read:lists`` permission.

        :param list_id: List ID.
        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonAccountSchema(many=True)
        """
        return self._run(
            f'lists/{list_id}/accounts',
            version='v1',
            method='get',
            schema=MastodonAccountSchema(many=True),
            params={
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
            },
            **kwargs
        )

    @action
    def add_accounts_to_list(self, list_id: int, account_ids: Sequence[str], **kwargs):
        """
        Add accounts to a list.

        It requires the specified API token to have the ``write:lists`` permission.

        :param list_id: List ID.
        :param account_ids: Accounts that should be added.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonListSchema
        """
        return self._run(
            f'lists/{list_id}/accounts',
            version='v1',
            method='post',
            data={'account_ids': account_ids},
            **kwargs
        )

    @action
    def remove_accounts_from_list(self, list_id: int, account_ids: Sequence[str], **kwargs):
        """
        Remove accounts from a list

        It requires the specified API token to have the ``write:lists`` permission.

        :param list_id: List ID.
        :param account_ids: Accounts that should be removed.
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonListSchema
        """
        return self._run(
            f'lists/{list_id}/accounts',
            version='v1',
            method='delete',
            data={'account_ids': account_ids},
            **kwargs
        )

    @action
    def get_notifications(
            self, notification_id: Optional[str] = None, min_id: Optional[str] = None,
            max_id: Optional[str] = None, limit: int = 20, **kwargs
    ) -> Union[dict, Iterable[dict]]:
        """
        Get the list of notifications of the user.

        It requires the specified API token to have the ``read:notifications`` permission.

        :param notification_id: If specified then retrieve only the notification associated to this ID.
        :param min_id: Return results newer than this ID.
        :param max_id: Return results older than this ID.
        :param limit: Maximum number of results (default: 20).
        :param kwargs: ``base_url``/``access_token`` override.
        :return: .. schema:: mastodon.MastodonNotificationSchema(many=True)
        """
        rs = self._run(
            'notifications' + (f'/{notification_id}' if notification_id else ''),
            version='v1',
            method='get',
            schema=MastodonNotificationSchema(many=notification_id is None),
            params={
                **({'min_id': min_id} if min_id else {}),
                **({'max_id': max_id} if max_id else {}),
                **({'limit': limit} if limit else {}),
            },
            **kwargs
        )

        return rs

    @action
    def dismiss_notifications(self, notification_id: Optional[str] = None, **kwargs):
        """
        Dismiss notifications.

        It requires the specified API token to have the ``write:notifications`` permission.

        :param notification_id: Dismiss only this notification.
        :param kwargs: ``base_url``/``access_token`` override.
        """
        return self._run(
            'notifications/' + (
                f'{notification_id}/dismiss' if notification_id else 'clear'
            ),
            version='v1',
            method='post',
            **kwargs
        )


# vim:sw=4:ts=4:et:



from collections.abc import Collection
import time
from typing import List, Optional, Union

from platypush.context import get_bus
from platypush.entities.distance import DistanceSensor
from platypush.message.event.distance import DistanceSensorEvent
from platypush.plugins import action
from platypush.plugins.gpio import GpioPlugin
from platypush.plugins.sensor import SensorPlugin


class SensorHcsr04Plugin(GpioPlugin, SensorPlugin):
    """
    You can use this plugin to interact with a distance sensor on your
    Raspberry Pi. It's been tested with a `HC-SR04 ultrasound sensor
    <https://www.sparkfun.com/products/15569>`_, but it should be compatible
    with any GPIO-compatible sensor that relies on the same trigger-and-echo
    principle.
    """

    def __init__(
        self,
        trigger_pin: int,
        echo_pin: int,
        poll_interval: float = 0.25,
        timeout: float = 2.0,
        warmup_time: float = 2.0,
        *args,
        **kwargs,
    ):
        """
        :param trigger_pin: GPIO PIN where you connected your sensor trigger
            PIN (the one that triggers the sensor to perform a measurement).
        :param echo_pin: GPIO PIN where you connected your sensor echo PIN (the
            one that will listen for the signal to bounce back and therefore
            trigger the distance calculation).
        :param timeout: The echo-wait will terminate and the plugin will return
            null if no echo has been received after this time (default: 1
            second).
        :param warmup_time: Number of seconds that should be waited on plugin
            instantiation for the sensor to be ready (default: 2 seconds).
        """

        measurement_interval = kwargs.pop('measurement_interval', None)
        if measurement_interval is not None:
            self.logger.warning(
                'measurement_interval is deprecated, use poll_interval instead',
            )
            poll_interval = measurement_interval

        SensorPlugin.__init__(self, poll_interval=poll_interval, *args, **kwargs)
        GpioPlugin.__init__(
            self,
            pins={
                'trigger': trigger_pin,
                'echo': echo_pin,
            },
            *args,
            **kwargs,
        )

        self.trigger_pin = trigger_pin
        self.echo_pin = echo_pin
        self.timeout = timeout
        self.warmup_time = warmup_time
        self._init_board()

    def _init_board(self):
        import RPi.GPIO as GPIO

        with self._init_lock:
            if self._initialized:
                return

            GpioPlugin._init_board(self)
            self._initialized = False

            GPIO.setup(self.trigger_pin, GPIO.OUT)
            GPIO.setup(self.echo_pin, GPIO.IN)
            GPIO.output(self.trigger_pin, GPIO.LOW)

            self.logger.info(
                'Waiting {} seconds for the sensor to be ready'.format(self.warmup_time)
            )
            self.wait_stop(self.warmup_time)
            self.logger.info('Sensor ready')
            self._initialized = True

    def _get_data(self):
        import RPi.GPIO as GPIO

        pulse_start = pulse_on = time.time()

        self._init_board()
        GPIO.output(self.trigger_pin, GPIO.HIGH)
        time.sleep(0.00001)  # 1 us pulse to trigger echo measurement
        GPIO.output(self.trigger_pin, GPIO.LOW)

        while GPIO.input(self.echo_pin) == 0:
            pulse_on = time.time()
            if pulse_on - pulse_start > self.timeout:
                raise TimeoutError(
                    'Distance sensor echo timeout after {} seconds: 0'.format(
                        self.timeout
                    )
                )

        pulse_start = pulse_on
        pulse_end = pulse_off = time.time()

        while GPIO.input(self.echo_pin) == 1:
            pulse_off = time.time()
            if pulse_off - pulse_end > self.timeout:
                raise TimeoutError(
                    'Distance sensor echo timeout after {} seconds: 1'.format(
                        self.timeout
                    )
                )

        pulse_end = pulse_off
        pulse_duration = pulse_end - pulse_start

        # s = vt where v = 1/2 * avg speed of sound in mm/s
        return round(pulse_duration * 171500.0, 2)

    @action
    def get_measurement(self, *_, **__) -> Optional[float]:
        """
        :returns: Distance measurement as a scalar (in mm):
        """

        try:
            distance = self._get_data()
            bus = get_bus()
            bus.post(DistanceSensorEvent(distance=distance, unit='mm'))
            return distance
        except TimeoutError as e:
            self.logger.warning(f'Read timeout: {e}')
            return None
        except Exception as e:
            self.cleanup()
            raise e

    def transform_entities(
        self, entities: Union[Optional[float], Collection[Optional[float]]]
    ) -> List[DistanceSensor]:
        value = (
            entities[0]
            if isinstance(entities, (list, tuple)) and len(entities)
            else entities
        )

        if value is None:
            return []

        return [
            DistanceSensor(
                id='hcsr04',
                name='HC-SR04 Distance Sensor',
                value=value,
                unit='mm',
            )
        ]

    @action
    def stop(self):
        super().stop()
        self.cleanup()


# vim:sw=4:ts=4:et:



import json
import struct
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bluetooth_numbers import company, oui

from TheengsDecoder import decodeBLE, getAttribute, getProperties

from platypush.entities import Entity
from platypush.entities.batteries import Battery
from platypush.entities.bluetooth import BluetoothDevice, BluetoothService
from platypush.entities.contact import ContactSensor
from platypush.entities.electricity import (
    CurrentSensor,
    EnergySensor,
    PowerSensor,
    VoltageSensor,
)
from platypush.entities.heart import HeartRateSensor
from platypush.entities.humidity import DewPointSensor, HumiditySensor
from platypush.entities.illuminance import IlluminanceSensor
from platypush.entities.motion import MotionSensor
from platypush.entities.presence import PresenceSensor
from platypush.entities.pressure import PressureSensor
from platypush.entities.sensors import BinarySensor, NumericSensor, RawSensor
from platypush.entities.steps import StepsSensor
from platypush.entities.temperature import TemperatureSensor
from platypush.entities.time import TimeDuration
from platypush.entities.weight import WeightSensor

from .._model import Protocol, ServiceClass


@dataclass
class TheengsEntity:
    """
    Utility class to store the data parsed from the Theengs library.
    """

    data: dict = field(default_factory=dict)
    properties: dict = field(default_factory=dict)
    manufacturer: Optional[str] = None
    model: Optional[str] = None
    model_id: Optional[str] = None


# pylint: disable=too-few-public-methods
class NullSensor:
    """
    Dummy class to model sensors with null values (hence without sufficient
    information for the application to infer the type).
    """

    def __init__(self, *_, **__):
        pass


# Maps property names to transformer methods (first mapper choice).
_property_to_entity: Dict[str, Callable[[Any, Dict[str, Any]], Entity]] = {
    'activity heart rate': lambda value, _: HeartRateSensor(value=value),
    'atmospheric pressure': lambda value, conf: PressureSensor(
        value=value,
        unit=conf.get('unit'),
    ),
    'battery': lambda value, conf: Battery(
        value=value,
        unit=conf.get('unit', '%'),
        min=conf.get('min', 0),
        max=conf.get('min', 100),
    ),
    'contact': lambda value, _: ContactSensor(value=value),
    'current': lambda value, conf: CurrentSensor(
        value=value,
        unit=conf.get('unit', 'A'),
    ),
    'dew_point_sensor': lambda value, conf: DewPointSensor(
        value=value,
        unit=conf.get('unit'),
    ),
    'duration': lambda value, conf: TimeDuration(
        value=value,
        unit=conf.get('unit'),
    ),
    'energy': lambda value, conf: EnergySensor(
        value=value,
        unit=conf.get('unit', 'kWh'),
    ),
    'heart rate': lambda value, _: HeartRateSensor(value=value),
    'humidity': lambda value, conf: HumiditySensor(
        value=value,
        unit=conf.get('unit', '%'),
        min=conf.get('min', 0),
        max=conf.get('min', 100),
    ),
    'light level': lambda value, conf: IlluminanceSensor(
        value=value,
        unit=conf.get('unit'),
    ),
    'luminance': lambda value, conf: IlluminanceSensor(
        value=value,
        unit=conf.get('unit'),
    ),
    'moisture': lambda value, conf: HumiditySensor(
        value=value,
        unit=conf.get('unit'),
    ),
    'motion': lambda value, _: MotionSensor(value=value),
    'open': lambda value, _: BinarySensor(value=value),
    'power': lambda value, conf: PowerSensor(
        value=value,
        unit=conf.get('unit', 'W'),
    ),
    'presence': lambda value, _: PresenceSensor(value=value),
    'pressure': lambda value, conf: PressureSensor(
        value=value,
        unit=conf.get('unit'),
    ),
    'steps': lambda value, _: StepsSensor(value=value),
    'temperature': lambda value, conf: TemperatureSensor(
        value=value,
        unit=conf.get('unit', 'C'),
    ),
    'temperature2': lambda value, conf: TemperatureSensor(
        value=value,
        unit=conf.get('unit', 'C'),
    ),
    'temperature3': lambda value, conf: TemperatureSensor(
        value=value,
        unit=conf.get('unit', 'C'),
    ),
    'temperature4': lambda value, conf: TemperatureSensor(
        value=value,
        unit=conf.get('unit', 'C'),
    ),
    'temperature5': lambda value, conf: TemperatureSensor(
        value=value,
        unit=conf.get('unit', 'C'),
    ),
    'temperature6': lambda value, conf: TemperatureSensor(
        value=value,
        unit=conf.get('unit', 'C'),
    ),
    'temperature7': lambda value, conf: TemperatureSensor(
        value=value,
        unit=conf.get('unit', 'C'),
    ),
    'temperature8': lambda value, conf: TemperatureSensor(
        value=value,
        unit=conf.get('unit', 'C'),
    ),
    'volt': lambda value, conf: VoltageSensor(
        value=value,
        unit=conf.get('unit', 'V'),
    ),
    'voltage': lambda value, conf: VoltageSensor(
        value=value,
        unit=conf.get('unit', 'V'),
    ),
    'weight': lambda value, conf: WeightSensor(
        value=value,
        unit=conf.get('unit', 'kg'),
    ),
}

# Maps reported units to transformer methods (second mapper choice).
_unit_to_entity: Dict[str, Callable[[Any, Dict[str, Any]], Entity]] = {
    'status': lambda value, _: BinarySensor(value=value),
    'int': lambda value, _: NumericSensor(value=value),
    'float': lambda value, _: NumericSensor(value=value),
    '%': lambda value, conf: NumericSensor(
        value=value,
        unit=conf.get('unit', '%'),
        min=conf.get('min', 0),
        max=conf.get('min', 100),
    ),
}


# Maps value types to transformer methods (third mapper choice).
_value_type_to_entity: Dict[type, Callable[[Any, Dict[str, Any]], Entity]] = {
    bool: lambda value, _: BinarySensor(value=value),
    int: lambda value, _: NumericSensor(value=value),
    float: lambda value, _: NumericSensor(value=value),
    str: lambda value, _: RawSensor(value=value),
    bytes: lambda value, _: RawSensor(value=value),
    bytearray: lambda value, _: RawSensor(value=value),
}


def _parse_services(
    device: BLEDevice, data: AdvertisementData
) -> List[BluetoothService]:
    """
    :param device: The target device.
    :param data: Published beacon data.
    :return: The parsed BLE services as a list of
        :class:`platypush.entities.bluetooth.BluetoothService`.
    """
    services: List[BluetoothService] = []
    for srv in data.service_uuids or []:
        try:
            uuid = BluetoothService.to_uuid(srv)
        except (TypeError, ValueError):
            # Not a valid UUID.
            continue

        srv_cls = ServiceClass.get(uuid)
        services.append(
            BluetoothService(
                id=f'{device.address}::{uuid}',
                uuid=uuid,
                name=f'[{uuid}]' if srv_cls == ServiceClass.UNKNOWN else str(srv_cls),
                protocol=Protocol.L2CAP,
                is_ble=True,
            )
        )

    return services


def device_to_entity(device: BLEDevice, data: AdvertisementData) -> BluetoothDevice:
    """
    Convert the data received from a Bluetooth advertisement packet into a
    compatible Platypush :class:`platypush.entity.bluetooth.BluetoothDevice`
    entity, with the discovered services and characteristics exposed as children
    entities.
    """

    theengs_entity = _parse_advertisement_data(data)
    props = (device.details or {}).get('props', {})
    manufacturer = _parse_manufacturer(device, theengs_entity, data)
    parent_entity = BluetoothDevice(
        id=device.address,
        model=theengs_entity.model,
        reachable=True,
        supports_ble=True,
        supports_legacy=False,
        address=device.address,
        name=device.name or device.address,
        connected=props.get('Connected', False),
        rssi=data.rssi,
        tx_power=props.get('TxPower'),
        manufacturer=manufacturer,
        children=_parse_services(device, data),
    )

    parsed_entities = {
        # Check if we can infer an entity mapper from the property name.
        conf.get('name', name): _property_to_entity.get(
            conf.get('name'),
            # If not, check if we can infer an entity mapper from the reported unit.
            _unit_to_entity.get(
                conf.get('unit'),
                # If not, check if we can infer an entity mapper from the value type.
                _value_type_to_entity.get(
                    type(theengs_entity.data.get(name)),
                    # If not, default to a NullSensor.
                    lambda *_: NullSensor(),
                ),
            ),
        )(theengs_entity.data.get(name), conf)
        for name, conf in theengs_entity.properties.items()
    }

    for prop, entity in parsed_entities.items():
        if isinstance(entity, NullSensor):
            # Skip entities that we couldn't parse.
            continue

        entity.id = f'{parent_entity.address}::{prop}'
        entity.name = prop.title()
        parent_entity.children.append(entity)
        entity.parent = parent_entity

    return parent_entity


def _parse_manufacturer(
    device: BLEDevice, entity: TheengsEntity, data: AdvertisementData
) -> Optional[str]:
    """
    :param device: The target device.
    :param entity: The entity that maps the received beacon data.
    :param data:
    :return: The parsed manufacturer name.
    """

    # If the manufacturer has already been parsed, return it.
    if entity.manufacturer:
        return entity.manufacturer

    # Otherwise, infer it from the first three bytes of the MAC address.
    manufacturer = oui.get(':'.join(device.address.split(':')[:3]).upper())
    if manufacturer:
        return manufacturer

    # Otherwise, infer it from the reported manufacturer_data.
    for key in data.manufacturer_data:
        manufacturer = company.get(key)
        if manufacturer:
            return manufacturer

    # Otherwise, we couldn't parse the manufacturer.
    return None


def _parse_advertisement_data(data: AdvertisementData) -> TheengsEntity:
    """
    :param data: The data received from a Bluetooth advertisement packet.
    :return: A :class:`platypush.entity.bluetooth.TheengsEntity` instance that
        maps the parsed attributes.
    """

    entity_args, properties, manufacturer, model, model_id = ({}, {}, None, None, None)

    if data.service_data:
        parsed_data = list(data.service_data.keys())[0]
        # TheengsDecoder only accepts 16 bit uuid's, this converts the 128 bit uuid to 16 bit.
        entity_args['servicedatauuid'] = parsed_data[4:8]
        parsed_data = str(list(data.service_data.values())[0].hex())
        entity_args['servicedata'] = parsed_data

    if data.manufacturer_data:
        parsed_data = str(
            struct.pack('<H', list(data.manufacturer_data.keys())[0]).hex()
        )
        parsed_data += str(list(data.manufacturer_data.values())[0].hex())
        entity_args['manufacturerdata'] = parsed_data

    if data.local_name:
        entity_args['name'] = data.local_name

    if entity_args:
        encoded_ret = decodeBLE(json.dumps(entity_args))

        if encoded_ret:
            entity_args = json.loads(encoded_ret)

            if entity_args.get('model_id'):
                properties = json.loads(getProperties(entity_args['model_id'])).get(
                    'properties', {}
                )
                model = getAttribute(entity_args['model_id'], 'model')

        model_id = entity_args.pop('model_id', None)

    return TheengsEntity(
        data=entity_args,
        properties=properties,
        manufacturer=manufacturer,
        model=model,
        model_id=model_id,
    )



from abc import ABC, abstractmethod
import json
from logging import getLogger, Logger


class Command(ABC):
    """
    Base class for application commands.
    """

    END_OF_COMMAND = b'\x00'
    """End-of-command marker."""

    def __init__(self, **args) -> None:
        self.args = args

    @property
    def logger(self) -> Logger:
        """
        The command class logger.
        """
        return getLogger(self.__class__.__name__)

    @abstractmethod
    def __call__(self, app, *_, **__):
        """
        Execute the command.
        """
        raise NotImplementedError()

    def __str__(self) -> str:
        """
        :return: A JSON representation of the command.
        """
        return json.dumps(
            {
                'type': 'command',
                'command': self.__class__.__name__,
                'args': self.args,
            }
        )

    def to_bytes(self):
        """
        :return: A JSON representation of the command.
        """
        return str(self).encode('utf-8') + self.END_OF_COMMAND

    @classmethod
    def parse(cls, data: bytes) -> "Command":
        """
        :param data: A JSON representation of the command.
        :raise ValueError: If the data is invalid.
        :return: The command instance or None if the data is invalid.
        """
        import platypush.commands

        try:
            json_data = json.loads(data.decode('utf-8'))
        except json.JSONDecodeError as e:
            raise ValueError from e

        kind = json_data.pop('type', None)
        if kind != 'command':
            raise ValueError(f'Invalid command type: {kind}')

        command_name = json_data.get('command')
        if not command_name:
            raise ValueError(f'Invalid command name: {command_name}')

        cmd_class = getattr(platypush.commands, command_name, None)
        if not (cmd_class and issubclass(cmd_class, Command)):
            raise ValueError(f'Invalid command class: {command_name}')

        try:
            return cmd_class(**json_data.get('args', {}))
        except Exception as e:
            raise ValueError(e) from e



from contextlib import contextmanager
import sys
from threading import RLock
from typing import Collection, Generator, Optional, Dict, Any, List, Union

from sqlalchemy.orm import Session

from platypush.context import get_plugin
from platypush.entities import EntityManager
from platypush.entities.alarm import Alarm as DbAlarm
from platypush.message.event.entities import EntityDeleteEvent
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.db import DbPlugin
from platypush.plugins.media import MediaPlugin
from platypush.procedure import Procedure
from platypush.utils import get_plugin_name_by_class
from platypush.utils.media import get_default_media_plugin

from ._model import Alarm, AlarmState


class AlarmPlugin(RunnablePlugin, EntityManager):
    """
    Alarm/timer plugin.

    It requires at least one enabled ``media`` plugin to be configured if you
    want to play audio resources.

    Example configuration:

    .. code-block:: yaml

        alarm:
            # Media plugin that will be used to play the alarm audio.
            # If not specified, the first available configured media plugin
            # will be used.
            media_plugin: media.vlc

            alarms:
                morning_alarm:
                    # Cron expression format: run every weekday at 7 AM
                    when: '0 7 * * 1-5'
                    media: ~/path/your_ringtone.mp3
                    audio_volume: 10       # 10%

                    # Repeat the played media resource until the alarm is
                    # snoozed/dismissed (default: true)
                    media_repeat: true

                    # Wait 5 minutes between a snooze and another run
                    snooze_interval: 300

                    # After 10 minutes with no manual snooze/dismiss,
                    # stop the alarm
                    dismiss_interval: 600

                    # Actions to be executed when the alarm goes on
                    actions:
                        - action: tts.say
                          args:
                              text: Good morning

                        - action: light.hue.bri
                          args:
                              value: 1

                        - action: light.hue.bri
                          args:
                              value: 140
                              transitiontime: 150

                one_shot_alarm:
                    # One-shot execution, with timestamp in ISO format
                    when: '2020-02-18T07:00:00.000000'
                    media: ~/path/your_ringtone.mp3
                    actions:
                        - action: light.hue.on

                timer:
                    # This alarm will execute the specified number of seconds
                    # after being initialized (5 minutes after the plugin has
                    # been initialized in this case)
                    when: 300
                    media: ~/path/your_ringtone.mp3
                    actions:
                        - action: light.hue.on

    """

    def __init__(
        self,
        alarms: Optional[Union[List[dict], Dict[str, dict]]] = None,
        media_plugin: Optional[str] = None,
        poll_interval: Optional[float] = 2.0,
        snooze_interval: float = 300.0,
        dismiss_interval: float = 300.0,
        **kwargs,
    ):
        """
        :param alarms: List or name->value dict with the configured alarms.
        :param media_plugin: Media plugin (instance of
            :class:`platypush.plugins.media.MediaPlugin`) that will be used to
            play the alarm audio. It needs to be a supported local media
            plugin, e.g. ``media.mplayer``, ``media.vlc``, ``media.mpv``,
            ``media.gstreamer``, ``sound``, etc. If not specified, the first
            available configured local media plugin will be used. This only
            applies to alarms that are configured to play an audio resource.
        :param poll_interval: (Internal) poll interval, in seconds (default: 2).
        :param snooze_interval: Default snooze interval in seconds. This
            specifies how long to wait between alarm runs when an alarm is
            dismissed (default: 300).
        :param dismiss_interval: Default dismiss interval in seconds. This
            specifies how long an alarm should run without being manually
            snoozed/dismissed before being automatically dismissed (default:
            300).
        """
        super().__init__(poll_interval=poll_interval, **kwargs)
        self.snooze_interval = snooze_interval
        self.dismiss_interval = dismiss_interval
        self._db_lock = RLock()
        alarms = alarms or []
        if isinstance(alarms, dict):
            alarms = [{'name': name, **alarm} for name, alarm in alarms.items()]

        if kwargs.get('audio_plugin'):
            self.logger.warning(
                'The audio_plugin parameter is deprecated. Use media_plugin instead'
            )
            media_plugin = media_plugin or kwargs.get('audio_plugin')

        try:
            plugin: Optional[MediaPlugin] = (
                get_plugin(media_plugin) if media_plugin else get_default_media_plugin()
            )
            assert plugin, 'No media/audio plugin configured'
            self.media_plugin = get_plugin_name_by_class(plugin.__class__)
        except AssertionError:
            self.media_plugin = None
            self.logger.warning(
                'No media plugin configured. Alarms that require audio playback will not work'
            )

        self.alarms = {
            alarm.name: alarm
            for alarm in [
                Alarm(
                    stop_event=self._should_stop,
                    static=True,
                    media_plugin=alarm.pop('media_plugin', self.media_plugin),
                    on_change=self._on_alarm_update,
                    **alarm,
                )
                for alarm in alarms
            ]
        }

        self._synced = False

    @property
    def _db(self) -> DbPlugin:
        db = get_plugin('db')
        assert db, 'No database plugin configured'
        return db

    @contextmanager
    def _get_session(self) -> Generator[Session, None, None]:
        with self._db_lock, self._db.get_session() as session:
            yield session

    def _merge_alarms(self, alarms: Dict[str, DbAlarm], session: Session):
        for name, alarm in alarms.items():
            if name in self.alarms:
                existing_alarm = self.alarms[name]

                # If the alarm is static, then we only want to override its
                # enabled state from the db record
                if existing_alarm.static:
                    existing_alarm.set_enabled(bool(alarm.enabled))
            else:
                # If the alarm record on the db is static, but the alarm is no
                # longer present in the configuration, then we want to delete it
                if alarm.static:
                    self._clear_alarm(alarm, session)
                else:
                    self.alarms[name] = Alarm.from_db(
                        alarm,
                        stop_event=self._should_stop,
                        media_plugin=alarm.media_plugin or self.media_plugin,
                        on_change=self._on_alarm_update,
                    )

        # Stop and remove alarms that are not statically configured no longer
        # present in the db
        for name, alarm in self.alarms.copy().items():
            if not alarm.static and name not in alarms:
                del self.alarms[name]
                alarm.stop()

    def _sync_alarms(self):
        with self._get_session() as session:
            db_alarms = {
                str(alarm.name): alarm for alarm in session.query(DbAlarm).all()
            }

            self._merge_alarms(db_alarms, session)
            self._clear_expired_alarms(session)
            for name, alarm in self.alarms.copy().items():
                if not (name in db_alarms or alarm.static):
                    self.alarms.pop(name, None)

        if not self._synced:
            self.publish_entities(self.alarms.values())
            self._synced = True

    def _clear_alarm(self, alarm: DbAlarm, session: Session):
        alarm_obj = self.alarms.pop(str(alarm.name), None)
        if alarm_obj:
            alarm_obj.stop()

        session.delete(alarm)
        self._bus.post(EntityDeleteEvent(entity=alarm))

    def _clear_expired_alarms(self, session: Session):
        expired_alarms = [
            alarm for alarm in self.alarms.values() if alarm.should_stop()
        ]

        if not expired_alarms:
            return

        expired_alarm_records = session.query(DbAlarm).filter(
            DbAlarm.name.in_([alarm.name for alarm in expired_alarms])
        )

        for alarm in expired_alarm_records:
            self._clear_alarm(alarm, session)

    def _get_alarms(self) -> List[Alarm]:
        return sorted(
            self.alarms.values(),
            key=lambda alarm: alarm.get_next() or sys.maxsize,
        )

    def _get_alarm(self, name: str) -> Alarm:
        assert name in self.alarms, f'The alarm {name} does not exist'
        return self.alarms[name]

    def _get_current_alarm(self) -> Optional[Alarm]:
        return next(
            iter(
                alarm
                for alarm in self.alarms.values()
                if alarm.state in {AlarmState.RUNNING, AlarmState.SNOOZED}
            ),
            None,
        )

    def _enable(self, name: str):
        self._get_alarm(name).enable()

    def _disable(self, name: str):
        self._get_alarm(name).disable()

    def _on_alarm_update(self, alarm: Alarm):
        with self._db_lock:
            if alarm.should_stop():
                return

            self.publish_entities([alarm])

    def _add(
        self,
        when: Union[str, int, float],
        actions: Union[list, Procedure],
        name: Optional[str] = None,
        media: Optional[str] = None,
        media_plugin: Optional[str] = None,
        media_repeat: bool = True,
        audio_file: Optional[str] = None,
        audio_volume: Optional[Union[int, float]] = None,
        enabled: bool = True,
        snooze_interval: Optional[float] = None,
        dismiss_interval: Optional[float] = None,
    ) -> Alarm:
        alarm = Alarm(
            when=when,
            actions=actions,
            name=name,
            enabled=enabled,
            media=media or audio_file,
            media_plugin=media_plugin or self.media_plugin,
            media_repeat=media_repeat,
            audio_volume=audio_volume,
            snooze_interval=snooze_interval or self.snooze_interval,
            dismiss_interval=dismiss_interval or self.dismiss_interval,
            stop_event=self._should_stop,
            on_change=self._on_alarm_update,
        )

        if alarm.name in self.alarms:
            assert not self.alarms[alarm.name].static, (
                f'Alarm {alarm.name} is statically defined in the configuration, '
                'cannot overwrite it programmatically'
            )

            self.logger.info('Overwriting existing alarm: %s', alarm.name)
            self.alarms[alarm.name].stop()

        self.alarms[alarm.name] = alarm
        alarm.start()
        self.publish_entities([alarm])
        return alarm

    def _dismiss(self):
        alarm = self._get_current_alarm()
        if not alarm:
            self.logger.info('No alarm is running')
            return

        alarm.dismiss()

    def _snooze(self, interval: Optional[float] = None):
        alarm = self._get_current_alarm()
        if not alarm:
            self.logger.info('No alarm is running')
            return

        interval = interval or alarm.snooze_interval or self.snooze_interval
        alarm.snooze(interval=interval)

    @action
    def add(
        self,
        when: Union[str, int, float],
        actions: Optional[list] = None,
        name: Optional[str] = None,
        media: Optional[str] = None,
        media_plugin: Optional[str] = None,
        media_repeat: bool = True,
        audio_file: Optional[str] = None,
        audio_volume: Optional[Union[int, float]] = None,
        enabled: bool = True,
        snooze_interval: Optional[float] = None,
        dismiss_interval: Optional[float] = None,
    ) -> dict:
        """
        Add a new alarm.

        :param when: When the alarm should be executed. It can be either a cron
            expression (for recurrent alarms), or a datetime string in ISO
            format (for one-shot alarms/timers), or an integer/float
            representing the number of seconds before the alarm goes on (e.g.
            300 for 5 minutes).
        :param actions: List of actions to be executed.
        :param name: Alarm name.
        :param media: Path of the audio file to be played.
        :param media_plugin: Override the default media plugin for this alarm.
        :param media_repeat: Repeat the played media resource until the alarm
            is snoozed/dismissed (default: True).
        :param audio_volume: Volume of the audio.
        :param enabled: Whether the new alarm should be enabled (default: True).
        :param snooze_interval: Snooze seconds before playing the alarm again.
        :param dismiss_interval: Dismiss seconds before stopping the alarm.
        :return: The newly created alarm.
        """
        if audio_file:
            self.logger.warning(
                'The audio_file parameter is deprecated. Use media instead'
            )

        return self._add(
            when=when,
            media=media,
            media_plugin=media_plugin,
            media_repeat=media_repeat,
            audio_file=audio_file,
            actions=actions or [],
            name=name,
            enabled=enabled,
            audio_volume=audio_volume,
            snooze_interval=snooze_interval,
            dismiss_interval=dismiss_interval,
        ).to_dict()

    @action
    def edit(
        self,
        name: str,
        new_name: Optional[str] = None,
        when: Optional[Union[str, int, float]] = None,
        actions: Optional[list] = None,
        media: Optional[str] = None,
        media_plugin: Optional[str] = None,
        media_repeat: Optional[bool] = None,
        audio_volume: Optional[Union[int, float]] = None,
        enabled: Optional[bool] = None,
        snooze_interval: Optional[float] = None,
        dismiss_interval: Optional[float] = None,
    ) -> dict:
        """
        Edit an existing alarm.

        Note that you can only edit the alarms that are not statically defined
        through the configuration.

        :param name: Alarm name.
        :param new_name: New alarm name.
        :param when: When the alarm should be executed. It can be either a cron
            expression (for recurrent alarms), or a datetime string in ISO
            format (for one-shot alarms/timers), or an integer/float
            representing the number of seconds before the alarm goes on (e.g.
            300 for 5 minutes).
        :param actions: List of actions to be executed.
        :param media: Path of the audio file to be played.
        :param media_plugin: Override the default media plugin for this alarm.
        :param media_repeat: Repeat the played media resource until the alarm
            is snoozed/dismissed (default: True).
        :param audio_volume: Volume of the audio.
        :param enabled: Whether the new alarm should be enabled.
        :param snooze_interval: Snooze seconds before playing the alarm again.
        :param dismiss_interval: Dismiss seconds before stopping the alarm.
        :return: The modified alarm.
        """
        alarm = self._get_alarm(name)
        assert not alarm.static, (
            f'Alarm {name} is statically defined in the configuration, '
            'cannot overwrite it programmatically'
        )

        if new_name and new_name != name:
            assert (
                new_name not in self.alarms
            ), f'An alarm with name {new_name} already exists'

        with self._db.get_session() as session:
            db_alarm = session.query(DbAlarm).filter_by(name=name).first()
            self._clear_alarm(db_alarm, session)
            return self._add(
                when=when or alarm.when,
                media=media or alarm.media,
                media_plugin=media_plugin or alarm.media_plugin or self.media_plugin,
                media_repeat=media_repeat
                if media_repeat is not None
                else alarm.media_repeat,
                actions=actions if actions is not None else (alarm.actions or []),
                name=new_name or name,
                enabled=enabled if enabled is not None else alarm.is_enabled(),
                audio_volume=audio_volume
                if audio_volume is not None
                else alarm.audio_volume,
                snooze_interval=snooze_interval or alarm.snooze_interval,
                dismiss_interval=dismiss_interval or alarm.dismiss_interval,
            ).to_dict()

    @action
    def delete(self, name: str):
        """
        Delete an alarm.

        :param name: Alarm name.
        """
        try:
            alarm = self._get_alarm(name)
        except AssertionError:
            self.logger.warning('Alarm %s does not exist', name)
            return

        assert not alarm.static, (
            f'Alarm {name} is statically defined in the configuration, '
            'cannot overwrite it programmatically'
        )

        alarm.stop()

        with self._db.get_session() as session:
            db_alarm = session.query(DbAlarm).filter_by(name=name).first()
            if not db_alarm:
                self.logger.warning('Alarm %s does not exist', name)
                return

            self._clear_alarm(db_alarm, session)

    @action
    def enable(self, name: str):
        """
        Enable an alarm.

        :param name: Alarm name.
        """
        self._enable(name)

    @action
    def disable(self, name: str):
        """
        Disable an alarm. This will prevent the alarm from executing until re-enabled or until the application
        is restarted.

        :param name: Alarm name.
        """
        self._disable(name)

    @action
    def set_enabled(self, name: str, enabled: bool):
        """
        Enable/disable an alarm.

        :param name: Alarm name.
        :param enabled: Whether the alarm should be enabled.
        """
        if enabled:
            self._enable(name)
        else:
            self._disable(name)

    @action
    def dismiss(self):
        """
        Dismiss the alarm that is currently running.
        """
        self._dismiss()

    @action
    def snooze(self, interval: Optional[float] = None):
        """
        Snooze the alarm that is currently running for the specified number of seconds.
        The alarm will stop and resume again later.

        :param interval: Snooze seconds before playing the alarm again (default: 300).
        """
        self._snooze(interval=interval)

    @action
    def get_alarms(self) -> List[Dict[str, Any]]:
        """
        Deprecated alias for :meth:`.status`.
        """
        self.logger.warning('get_alarms() is deprecated. Use status() instead')
        return self.status()  # type: ignore

    @action
    def status(self, *_, **__) -> List[Dict[str, Any]]:
        """
        Get the list of configured alarms and their status.

        :return: List of the alarms, sorted by next scheduled run. Example:

            .. code-block:: json

                [
                    {
                        "name": "Morning alarm",
                        "id": 1,
                        "when": "0 8 * * 1-5",
                        "next_run": "2023-12-06T08:00:00.000000",
                        "enabled": true,
                        "media": "/path/to/media.mp3",
                        "media_plugin": "media.vlc",
                        "media_repeat": true,
                        "audio_volume": 10,
                        "snooze_interval": 300,
                        "dismiss_interval": 300,
                        "actions": [
                            {
                                "action": "tts.say",
                                "args": {
                                    "text": "Good morning"
                                }
                            },
                            {
                                "action": "light.hue.on"
                            }
                        ],
                        "state": "RUNNING"
                    }
                ]

        """
        ret = [alarm.to_dict() for alarm in self._get_alarms()]
        self.publish_entities(self.alarms.values())
        return ret

    def transform_entities(self, entities: Collection[Alarm], **_) -> List[DbAlarm]:
        return [alarm.to_db() for alarm in entities]

    def main(self):
        self._sync_alarms()
        for alarm in self.alarms.values():
            alarm.start()

        while not self.should_stop():
            self._sync_alarms()
            self.wait_stop(self.poll_interval)

    def stop(self):
        for alarm in self.alarms.values():
            alarm.stop()

        super().stop()


# vim:sw=4:ts=4:et:



from enum import Enum
from typing import Any, Callable, List, Optional, Type, Union

from pysmartthings import Attribute, Capability, Command, DeviceEntity

from platypush.entities import Entity

from platypush.entities.audio import Muted, Volume
from platypush.entities.batteries import Battery
from platypush.entities.dimmers import Dimmer
from platypush.entities.electricity import EnergySensor, PowerSensor, VoltageSensor
from platypush.entities.humidity import HumiditySensor
from platypush.entities.illuminance import IlluminanceSensor
from platypush.entities.linkquality import LinkQuality
from platypush.entities.motion import MotionSensor
from platypush.entities.sensors import (
    BinarySensor,
    EnumSensor,
    CompositeSensor,
    NumericSensor,
)
from platypush.entities.switches import EnumSwitch, Switch
from platypush.entities.temperature import TemperatureSensor


class DeviceMapper:
    """
    The purpose of these objects is to map the capabilities of SmartThings
    devices to native Platypush entities.
    """

    def __init__(
        self,
        entity_type: Type[Entity],
        capability: str,
        attribute: str,
        value_type: Union[Type, Enum, str],
        set_command: Optional[Union[str, Callable[[Any], str]]] = None,
        get_value: Optional[Callable[[DeviceEntity], Any]] = None,
        set_value_args: Optional[Callable[..., Any]] = None,
        **kwargs
    ):
        self.entity_type = entity_type
        self.capability = capability
        self.set_command = set_command
        self.attribute = attribute
        self.value_type = value_type
        self.get_value = get_value if get_value else self._default_get_value
        self.values: List[str] = []
        self.entity_args = kwargs

        if isinstance(value_type, Enum):
            self.values = [v.name for v in entity_type]  # type: ignore

        self.set_value_args = (
            set_value_args if set_value_args else self._default_set_value_args
        )

    def _cast_value(self, value: Any) -> Any:
        if not isinstance(self.value_type, (str, Enum)):
            try:
                value = self.value_type(value)
            except Exception:
                # Set the value to None in case of cast errors
                value = None

        return value

    def _default_get_value(self, device: DeviceEntity) -> Any:
        if hasattr(device.status, self.attribute):
            value = getattr(device.status, self.attribute)
        else:
            value = device.status.attributes[self.attribute].value

        return self._cast_value(value)

    def _default_set_value_args(self, *values: Any) -> List[Any]:
        return [self._cast_value(v) for v in values]


device_mappers: List[DeviceMapper] = [
    # acceleration_sensor
    DeviceMapper(
        entity_type=BinarySensor,
        capability=Capability.acceleration_sensor,
        attribute=Attribute.acceleration,
        value_type=bool,
    ),
    # air_conditioner_fan_mode
    DeviceMapper(
        entity_type=EnumSwitch,
        capability=Capability.air_conditioner_fan_mode,
        attribute=Attribute.fan_mode,
        value_type=Attribute.supported_ac_fan_modes,
        set_command=Command.set_fan_mode,
    ),
    # air_conditioner_mode
    DeviceMapper(
        entity_type=EnumSwitch,
        capability=Capability.air_conditioner_mode,
        attribute=Attribute.air_conditioner_mode,
        value_type=Attribute.supported_ac_modes,
        set_command=Command.set_air_conditioner_mode,
    ),
    # air_quality_sensor
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.air_quality_sensor,
        attribute=Attribute.air_quality,
        value_type=int,
    ),
    # alarm
    DeviceMapper(
        entity_type=EnumSwitch,
        capability=Capability.alarm,
        attribute=Attribute.alarm,
        value_type=Enum('AlarmValues', ['both', 'off', 'siren', 'strobe']),
        set_command=lambda value: value,
        set_value_args=lambda *_: [],
    ),
    # audio_mute
    DeviceMapper(
        entity_type=Muted,
        capability=Capability.audio_mute,
        attribute=Attribute.mute,
        value_type=bool,
        set_command=lambda value: Command.mute if value else Command.unmute,
        set_value_args=lambda *_: [],
    ),
    # audio_volume
    DeviceMapper(
        entity_type=Volume,
        capability=Capability.audio_volume,
        attribute=Attribute.volume,
        value_type=int,
        set_command=Command.set_volume,
    ),
    # battery
    DeviceMapper(
        entity_type=Battery,
        capability=Capability.battery,
        attribute=Attribute.battery,
        value_type=int,
    ),
    # body_mass_index_measurement
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.body_mass_index_measurement,
        attribute=Attribute.body_weight_measurement,
        value_type=float,
    ),
    # body_weight_measurement
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.body_weight_measurement,
        attribute=Attribute.body_weight_measurement,
        value_type=float,
    ),
    # button
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.button,
        attribute=Attribute.button,
        value_type=Enum(
            'ButtonValues',
            [
                'pushed',
                'held',
                'double',
                'pushed_2x',
                'pushed_3x',
                'pushed_4x',
                'pushed_5x',
                'pushed_6x',
                'down',
                'down_2x',
                'down_3x',
                'down_4x',
                'down_5x',
                'down_6x',
                'down_hold',
                'up',
                'up_2x',
                'up_3x',
                'up_4x',
                'up_5x',
                'up_6x',
                'up_hold',
                'swipe_up',
                'swipe_down',
                'swipe_left',
                'swipe_right',
                'unknown',
            ],
        ),
    ),
    # carbon_dioxide_measurement
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.carbon_dioxide_measurement,
        attribute=Attribute.carbon_dioxide,
        value_type=float,
    ),
    # carbon_monoxide_detector
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.carbon_monoxide_detector,
        attribute=Attribute.carbon_monoxide,
        value_type=Enum(
            'CarbonMonoxideValues', ['clear', 'detected', 'tested', 'unknown']
        ),
    ),
    # carbon_monoxide_measurement
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.carbon_monoxide_measurement,
        attribute=Attribute.carbon_monoxide_level,
        value_type=float,
    ),
    # contact_sensor
    DeviceMapper(
        entity_type=BinarySensor,
        capability=Capability.contact_sensor,
        attribute=Attribute.contact,
        value_type=bool,
    ),
    # dishwasher_operating_state
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.dishwasher_operating_state,
        attribute=Attribute.machine_state,
        value_type=Attribute.supported_machine_states,
    ),
    # door_control
    DeviceMapper(
        entity_type=Switch,
        capability=Capability.door_control,
        attribute=Attribute.door,
        value_type=bool,
        get_value=lambda device: device.status.door in {'open', 'opening'},
        set_command=lambda value: Command.open if value else Command.close,
    ),
    # dryer_operating_state
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.dryer_operating_state,
        attribute=Attribute.machine_state,
        value_type=Attribute.supported_machine_states,
    ),
    # dust_sensor
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.dust_sensor,
        attribute=Attribute.dust_level,
        value_type=float,
    ),
    # energy_meter
    DeviceMapper(
        entity_type=EnergySensor,
        capability=Capability.energy_meter,
        attribute=Attribute.energy,
        value_type=float,
    ),
    # equivalent_carbon_dioxide_measurement
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.equivalent_carbon_dioxide_measurement,
        attribute=Attribute.equivalent_carbon_dioxide_measurement,
        value_type=float,
    ),
    # fan_speed
    DeviceMapper(
        entity_type=Dimmer,
        capability=Capability.fan_speed,
        attribute=Attribute.fan_speed,
        value_type=int,
        set_command=Command.set_fan_speed,
    ),
    # formaldehyde_measurement
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.formaldehyde_measurement,
        attribute=Attribute.formaldehyde_level,
        value_type=float,
    ),
    # gas_meter
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.gas_meter,
        attribute=Attribute.gas_meter,
        value_type=float,
    ),
    # illuminance_measurement
    DeviceMapper(
        entity_type=IlluminanceSensor,
        capability=Capability.illuminance_measurement,
        attribute=Attribute.illuminance,
        value_type=float,
    ),
    # infrared_level
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.infrared_level,
        attribute=Attribute.infrared_level,
        value_type=int,
        min=0,
        max=100,
    ),
    # lock
    DeviceMapper(
        entity_type=Switch,
        capability=Capability.lock,
        attribute=Attribute.lock,
        value_type=bool,
        get_value=lambda device: device.status.lock in {True, 'locked'},
        set_command=lambda value: Command.lock if value else Command.unlock,
        set_value_args=lambda *_: [],
    ),
    # media_input_source
    DeviceMapper(
        entity_type=EnumSwitch,
        capability=Capability.media_input_source,
        attribute=Attribute.input_source,
        value_type=Attribute.supported_input_sources,
        set_command=Command.set_input_source,
    ),
    # motion_sensor
    DeviceMapper(
        entity_type=MotionSensor,
        capability=Capability.motion_sensor,
        attribute=Attribute.motion,
        value_type=bool,
    ),
    # odor_sensor
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.odor_sensor,
        attribute=Attribute.odor_level,
        value_type=int,
    ),
    # oven_operating_state
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.oven_operating_state,
        attribute=Attribute.machine_state,
        value_type=Attribute.supported_machine_states,
    ),
    # power_consumption_report
    DeviceMapper(
        entity_type=PowerSensor,
        capability=Capability.power_consumption_report,
        attribute=Attribute.power_consumption,
        value_type=float,
    ),
    # power_meter
    DeviceMapper(
        entity_type=PowerSensor,
        capability=Capability.power_meter,
        attribute=Attribute.power,
        value_type=float,
    ),
    # power_source
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.power_source,
        attribute=Attribute.power_source,
        value_type=Enum('PowerSourceValues', ['battery', 'dc', 'mains', 'unknown']),
    ),
    # presence_sensor
    DeviceMapper(
        entity_type=BinarySensor,
        capability=Capability.presence_sensor,
        attribute=Attribute.presence,
        value_type=bool,
    ),
    # relative_humidity_measurement
    DeviceMapper(
        entity_type=HumiditySensor,
        capability=Capability.relative_humidity_measurement,
        attribute=Attribute.humidity,
        value_type=int,
        min=0,
        max=100,
    ),
    # signal_strength
    DeviceMapper(
        entity_type=LinkQuality,
        capability=Capability.signal_strength,
        attribute=Attribute.lqi,
        value_type=int,
        min=0,
        max=255,
    ),
    # smoke_detector
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.smoke_detector,
        attribute=Attribute.smoke,
        value_type=Enum('SmokeDetectorValues', ['clear', 'detected', 'tested']),
    ),
    # sound_sensor
    DeviceMapper(
        entity_type=BinarySensor,
        capability=Capability.sound_sensor,
        attribute=Attribute.sound,
        value_type=bool,
    ),
    # switch
    DeviceMapper(
        entity_type=Switch,
        capability=Capability.switch,
        attribute=Attribute.switch,
        value_type=bool,
        set_command=lambda value: Command.on if value else Command.off,
        set_value_args=lambda *_: [],
    ),
    # switch_level
    DeviceMapper(
        entity_type=Dimmer,
        capability=Capability.switch_level,
        attribute=Attribute.level,
        value_type=int,
        set_command=Command.set_level,
    ),
    # tamper_alert
    DeviceMapper(
        entity_type=BinarySensor,
        capability=Capability.tamper_alert,
        attribute=Attribute.tamper,
        value_type=bool,
    ),
    # temperature_measurement
    DeviceMapper(
        entity_type=TemperatureSensor,
        capability=Capability.temperature_measurement,
        attribute=Attribute.temperature,
        value_type=float,
    ),
    # thermostat_cooling_setpoint
    DeviceMapper(
        entity_type=Dimmer,
        capability=Capability.thermostat_cooling_setpoint,
        attribute=Attribute.cooling_setpoint,
        value_type=float,
        min=-460,
        max=10000,
        set_command=Command.set_cooling_setpoint,
    ),
    # thermostat_fan_mode
    DeviceMapper(
        entity_type=EnumSwitch,
        capability=Capability.thermostat_fan_mode,
        attribute=Attribute.thermostat_fan_mode,
        value_type=Attribute.supported_thermostat_fan_modes,
        set_command=Command.set_thermostat_fan_mode,
    ),
    # thermostat_heating_setpoint
    DeviceMapper(
        entity_type=Dimmer,
        capability=Capability.thermostat_heating_setpoint,
        attribute=Attribute.heating_setpoint,
        value_type=float,
        min=-460,
        max=10000,
        set_command=Command.set_heating_setpoint,
    ),
    # thermostat_mode
    DeviceMapper(
        entity_type=EnumSwitch,
        capability=Capability.thermostat_mode,
        attribute=Attribute.thermostat_mode,
        value_type=Attribute.supported_thermostat_modes,
        set_command=Command.set_thermostat_mode,
    ),
    # thermostat_operating_state
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.thermostat_operating_state,
        attribute=Attribute.thermostat_operating_state,
        value_type=Enum(
            'ThermostatOperatingState',
            [
                'cooling',
                'fan only',
                'heating',
                'idle',
                'pending cool',
                'pending heat',
                'vent economizer',
            ],
        ),
    ),
    # three_axis
    DeviceMapper(
        entity_type=CompositeSensor,
        capability=Capability.three_axis,
        attribute=Attribute.three_axis,
        value_type=list,
    ),
    # tv_channel
    DeviceMapper(
        entity_type=Dimmer,
        capability=Capability.tv_channel,
        attribute=Attribute.tv_channel,
        value_type=int,
        set_command=Command.set_tv_channel,
        min=1,
        max=1000,
    ),
    # tvoc_measurement
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.tvoc_measurement,
        attribute=Attribute.tvoc_level,
        value_type=float,
    ),
    # ultraviolet_index
    DeviceMapper(
        entity_type=NumericSensor,
        capability=Capability.ultraviolet_index,
        attribute=Attribute.ultraviolet_index,
        value_type=float,
    ),
    # valve
    DeviceMapper(
        entity_type=Switch,
        capability=Capability.valve,
        attribute=Attribute.valve,
        value_type=bool,
        set_command=lambda value: Command.open if value else Command.close,
        set_value_args=lambda *_: [],
    ),
    # voltage_measurement
    DeviceMapper(
        entity_type=VoltageSensor,
        capability=Capability.voltage_measurement,
        attribute=Attribute.voltage,
        value_type=float,
    ),
    # washer_operating_state
    DeviceMapper(
        entity_type=EnumSensor,
        capability=Capability.washer_operating_state,
        attribute=Attribute.machine_state,
        value_type=Attribute.supported_machine_states,
    ),
    # water_sensor
    DeviceMapper(
        entity_type=BinarySensor,
        capability=Capability.water_sensor,
        attribute=Attribute.water,
        value_type=bool,
    ),
    # window_shade
    DeviceMapper(
        entity_type=Dimmer,
        capability=Capability.window_shade,
        attribute=Attribute.window_shade,
        value_type=int,
        set_command='setWindowShade',
        min=0,
        max=100,
    ),
]


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import Optional, Tuple

import numpy as np

from ._parser import SoundParser


class SoundBase(SoundParser, ABC):
    """
    Base class for synthetic sounds and mixes.
    """

    def __init__(self, *args, volume: float = 100, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.volume = volume

    @property
    def gain(self) -> float:
        return self.volume / 100

    @gain.setter
    def gain(self, value: float):
        self.volume = value * 100

    @abstractmethod
    def get_wave(
        self,
        sample_rate: float,
        t_start: float = 0,
        t_end: float = 0,
        **_,
    ):  # -> NDArray[np.floating]:
        """
        Get the wave binary data associated to this sound

        :param t_start: Start offset for the wave in seconds. Default: 0
        :param t_end: End offset for the wave in seconds. Default: 0
        :param sample_rate: Audio sample rate. Default: 44100 Hz
        :returns: A ``numpy.ndarray[(t_end-t_start)*sample_rate, 1]``
            with the raw float values
        """
        raise NotImplementedError()

    def fft(
        self,
        sample_rate: float,
        t_start: float = 0.0,
        t_end: float = 0.0,
        freq_range: Optional[Tuple[float, float]] = None,
        freq_buckets: Optional[int] = None,
    ):  # -> NDArray[np.floating]:
        """
        Get the real part of the Fourier transform associated to a time-bounded
            sample of this sound.

        :param t_start: Start offset for the wave in seconds. Default: 0
        :param t_end: End offset for the wave in seconds. Default: 0
        :param sample_rate: Audio sample rate. Default: 44100 Hz
        :param freq_range: FFT frequency range. Default: ``(0, sample_rate/2)``
            (see`Nyquist-Shannon sampling theorem
            <https://en.wikipedia.org/wiki/Nyquist%E2%80%93Shannon_sampling_theorem>`_)
        :param freq_buckets: Number of buckets to subdivide the frequency range.
            Default: None
        :returns: A numpy.ndarray[freq_range,1] with the raw float values
        """

        if not freq_range:
            freq_range = (0, int(sample_rate / 2))

        wave = self.get_wave(t_start=t_start, t_end=t_end, sample_rate=sample_rate)
        fft = np.fft.fft(wave.reshape(len(wave)))
        fft = fft.real[freq_range[0] : freq_range[1]]

        if freq_buckets is not None:
            fft = np.histogram(fft, bins=freq_buckets)[0]

        return fft



from marshmallow import fields
from marshmallow.schema import Schema
from marshmallow.validate import OneOf


class DbusSignalSchema(Schema):
    bus = fields.String(
        required=True,
        validate=OneOf(['system', 'session'])
    )

    interface = fields.String(allow_none=True, metadata={
        'description': 'The DBus interface that should be monitored (default: all)'
    })

    path = fields.String(allow_none=True, metadata={
        'description': 'Path of the resource to be monitored (default: all)'
    })

    signal = fields.String(allow_none=True, metadata={
        'description': 'Signal name filter (default: all signals)'
    })

    sender = fields.String(allow_none=True, metadata={
        'description': 'Signal sender filter (default: all senders)'
    })



from typing import Optional

from platypush.message.event import Event


class LinodeEvent(Event):
    """
    Base Linode event class.
    """


class LinodeInstanceStatusChanged(LinodeEvent):
    """
    Event triggered when the status of a Linode instance changes.
    """

    def __init__(
        self,
        *args,
        instance_id: int,
        instance_name: str,
        status: str,
        old_status: Optional[str] = None,
        **kwargs
    ):
        """
        :param instance_id: Linode instance ID.
        :param instance: Linode instance name.
        :param status: New status of the instance.
        :param old_status: Old status of the instance.
        """
        super().__init__(
            *args,
            instance_id=instance_id,
            instance_name=instance_name,
            status=status,
            old_status=old_status,
            **kwargs
        )


# vim:sw=4:ts=4:et:



from typing import Optional

from PIL import Image
from PIL.Image import Image as ImageType

from platypush.plugins.camera import CameraPlugin
from platypush.plugins.camera.gstreamer.model import GStreamerCamera
from platypush.common.gstreamer import Pipeline


class CameraGstreamerPlugin(CameraPlugin):
    """
    Plugin to interact with a camera over GStreamer.
    """

    _camera_class = GStreamerCamera

    def __init__(self, device: Optional[str] = '/dev/video0', **opts):
        """
        :param device: Path to the camera device (default ``/dev/video0``).
        :param opts: Camera options - see constructor of :class:`platypush.plugins.camera.CameraPlugin`.
        """
        super().__init__(device=device, **opts)

    def prepare_device(self, camera: GStreamerCamera, **_) -> Pipeline:
        pipeline = Pipeline()
        src = pipeline.add_source('v4l2src', device=camera.info.device)
        convert = pipeline.add('videoconvert')
        assert camera.info and camera.info.resolution

        video_filter = pipeline.add(
            'capsfilter',
            caps='video/x-raw,format=RGB,width={width},height={height},framerate={fps}/1'.format(
                width=camera.info.resolution[0],
                height=camera.info.resolution[1],
                fps=camera.info.fps,
            ),
        )

        sink = pipeline.add_sink('appsink', name='appsink', sync=False)
        pipeline.link(src, convert, video_filter, sink)
        return pipeline

    def start_camera(
        self, camera: GStreamerCamera, preview: bool = False, *args, **kwargs
    ):
        super().start_camera(*args, camera=camera, preview=preview, **kwargs)
        if camera.object:
            camera.object.play()

    def release_device(self, camera: GStreamerCamera):
        if camera.object:
            camera.object.stop()

    def capture_frame(self, camera: GStreamerCamera, *_, **__) -> Optional[ImageType]:
        if not (camera.info and camera.info.fps and camera.info.resolution):
            return None

        timed_out = not camera.object.data_ready.wait(
            timeout=5 + (1.0 / camera.info.fps)
        )
        if timed_out:
            self.logger.warning('Frame capture timeout')
            return None

        data = camera.object.data
        if data is None:
            return None

        camera.object.data_ready.clear()
        if (
            not data
            and len(data) != camera.info.resolution[0] * camera.info.resolution[1] * 3
        ):
            return None

        return Image.frombytes('RGB', camera.info.resolution, data)


# vim:sw=4:ts=4:et:



from marshmallow_dataclass import class_schema

from ._base import MemoryStatsBaseSchema
from ._model import MemoryStats, SwapStats


MemoryStatsSchema = class_schema(MemoryStats, base_schema=MemoryStatsBaseSchema)
SwapStatsSchema = class_schema(SwapStats, base_schema=MemoryStatsBaseSchema)



from marshmallow import Schema, fields, pre_dump, post_dump

from platypush.schemas import DateTime


class TidalSchema(Schema):
    pass


class TidalArtistSchema(TidalSchema):
    id = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'example': '3288612',
            'description': 'Artist ID',
        },
    )

    url = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'description': 'Artist Tidal URL',
            'example': 'https://tidal.com/artist/3288612',
        },
    )

    name = fields.String(required=True)

    @pre_dump
    def _prefill_url(self, data, *_, **__):
        data.url = f'https://tidal.com/artist/{data.id}'
        return data


class TidalAlbumSchema(TidalSchema):
    def __init__(self, *args, with_tracks=False, **kwargs):
        super().__init__(*args, **kwargs)
        self._with_tracks = with_tracks

    id = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'example': '45288612',
            'description': 'Album ID',
        },
    )

    url = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'description': 'Album Tidal URL',
            'example': 'https://tidal.com/album/45288612',
        },
    )

    name = fields.String(required=True)
    artist = fields.Nested(TidalArtistSchema)
    duration = fields.Int(metadata={'description': 'Album duration, in seconds'})
    year = fields.Integer(metadata={'example': 2003})
    num_tracks = fields.Int(metadata={'example': 10})
    tracks = fields.List(fields.Dict(), attribute='_tracks')

    @pre_dump
    def _prefill_url(self, data, *_, **__):
        data.url = f'https://tidal.com/album/{data.id}'
        return data

    @pre_dump
    def _cache_tracks(self, data, *_, **__):
        if self._with_tracks:
            album_id = str(data.id)
            self.context[album_id] = {
                'tracks': data.tracks(),
            }

        return data

    @post_dump
    def _dump_tracks(self, data, *_, **__):
        if self._with_tracks:
            album_id = str(data['id'])
            ctx = self.context.pop(album_id, {})
            data['tracks'] = TidalTrackSchema().dump(ctx.pop('tracks', []), many=True)

        return data


class TidalTrackSchema(TidalSchema):
    id = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'example': '25288614',
            'description': 'Track ID',
        },
    )

    url = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'description': 'Track Tidal URL',
            'example': 'https://tidal.com/track/25288614',
        },
    )

    artist = fields.Nested(TidalArtistSchema)
    album = fields.Nested(TidalAlbumSchema)
    name = fields.String(metadata={'description': 'Track title'})
    duration = fields.Int(metadata={'description': 'Track duration, in seconds'})
    track_num = fields.Int(
        metadata={'description': 'Index of the track within the album'}
    )

    @pre_dump
    def _prefill_url(self, data, *_, **__):
        data.url = f'https://tidal.com/track/{data.id}'
        return data


class TidalPlaylistSchema(TidalSchema):
    id = fields.String(
        required=True,
        dump_only=True,
        attribute='uuid',
        metadata={
            'example': '2b288612-34f5-11ed-b42d-001500e8f607',
            'description': 'Playlist ID',
        },
    )

    url = fields.String(
        required=True,
        dump_only=True,
        metadata={
            'description': 'Playlist Tidal URL',
            'example': 'https://tidal.com/playlist/2b288612-34f5-11ed-b42d-001500e8f607',
        },
    )

    name = fields.String(required=True)
    description = fields.String()
    duration = fields.Int(metadata={'description': 'Playlist duration, in seconds'})
    public = fields.Boolean(attribute='publicPlaylist')
    owner = fields.String(
        attribute='creator',
        metadata={
            'description': 'Playlist creator/owner ID',
        },
    )

    num_tracks = fields.Int(
        attribute='numberOfTracks',
        default=0,
        metadata={
            'example': 42,
            'description': 'Number of tracks in the playlist',
        },
    )

    created_at = DateTime(
        attribute='created',
        metadata={
            'description': 'When the playlist was created',
        },
    )

    last_updated_at = DateTime(
        attribute='lastUpdated',
        metadata={
            'description': 'When the playlist was last updated',
        },
    )

    tracks = fields.Nested(TidalTrackSchema, many=True)

    def _flatten_object(self, data, *_, **__):
        if not isinstance(data, dict):
            data = {
                'created': data.created,
                'creator': data.creator.id,
                'description': data.description,
                'duration': data.duration,
                'lastUpdated': data.last_updated,
                'uuid': data.id,
                'name': data.name,
                'numberOfTracks': data.num_tracks,
                'publicPlaylist': data.public,
                'tracks': getattr(data, '_tracks', []),
            }

        return data

    def _normalize_owner(self, data, *_, **__):
        owner = data.pop('owner', data.pop('creator', None))
        if owner:
            if isinstance(owner, dict):
                owner = owner['id']
            data['creator'] = owner

        return data

    def _normalize_name(self, data, *_, **__):
        if data.get('title'):
            data['name'] = data.pop('title')
        return data

    @pre_dump
    def normalize(self, data, *_, **__):
        if not isinstance(data, dict):
            data = self._flatten_object(data)

        self._normalize_name(data)
        self._normalize_owner(data)
        if 'tracks' not in data:
            data['tracks'] = []
        return data


class TidalSearchResultsSchema(TidalSchema):
    artists = fields.Nested(TidalArtistSchema, many=True)
    albums = fields.Nested(TidalAlbumSchema, many=True)
    tracks = fields.Nested(TidalTrackSchema, many=True)
    playlists = fields.Nested(TidalPlaylistSchema, many=True)



import base64
import mimetypes
import os

from email.encoders import encode_base64
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from platypush.plugins import action
from platypush.plugins.google import GooglePlugin


class GoogleMailPlugin(GooglePlugin):
    r"""
    GMail plugin. It allows you to programmatically compose and (TODO) get emails.

    To use this plugin:

        1. Create your Google application, if you don't have one already, on
           the `developers console <https://console.developers.google.com>`_.

        2. You may have to explicitly enable your user to use the app if the app
           is created in test mode. Go to "OAuth consent screen" and add your user's
           email address to the list of authorized users.

        3. Select the scopes that you want to enable for your application, depending
           on the integrations that you want to use.
           See https://developers.google.com/identity/protocols/oauth2/scopes
           for a list of the available scopes.

        4. Click on "Credentials", then "Create credentials" -> "OAuth client ID".

        5. Select "Desktop app", enter whichever name you like, and click "Create".

        6. Click on the "Download JSON" icon next to your newly created client ID.
           Save the JSON file under
           ``<WORKDIR>/credentials/google/client_secret.json``.

        7. If you're running the service on a desktop environment, then you
           can just start the application. A browser window will open and
           you'll be asked to authorize the application - you may be prompted
           with a warning because you are running a personal and potentially
           unverified application. After authorizing the application, the
           process will save the credentials under
           ``<WORKDIR>/credentials/google/<list,of,scopes>.json`` and proceed
           with the plugin initialization.

        8. If you're running the service on a headless environment, or you
           prefer to manually generate the credentials file before copying to
           another machine, you can run the following command:

                .. code-block:: bash

                  mkdir -p <WORKDIR>/credentials/google
                  python -m platypush.plugins.google.credentials \
                      'gmail.modify' \
                      <WORKDIR>/credentials/google/client_secret.json [--noauth_local_webserver]

           When launched with ``--noauth_local_webserver``, the script will
           start a local webserver and print a URL that should be opened in
           your browser. After authorizing the application, you may be
           prompted with a code that you should copy and paste back to the
           script. Otherwise, if you're running the script on a desktop, a
           browser window will be opened automatically.

    """

    scopes = ['https://www.googleapis.com/auth/gmail.modify']

    def __init__(self, *args, **kwargs):
        super().__init__(scopes=self.scopes, *args, **kwargs)

    @action
    def compose(self, sender, to, subject, body, files=None):
        """
        Compose a message.

        :param sender: Sender email/name
        :type sender: str

        :param to: Recipient email or comma-separated list of recipient emails
        :type to: str

        :param subject: Email subject
        :type subject: str

        :param body: Email body
        :type body: str

        :param files: Optional list of files to attach
        :type files: list
        """

        message = MIMEMultipart() if files else MIMEText(body)
        message['to'] = to
        message['from'] = sender
        message['subject'] = subject

        if files:
            for file in files:
                msg = MIMEText(body)
                message.attach(msg)
                content_type, encoding = mimetypes.guess_type(file)

                if content_type is None or encoding is not None:
                    content_type = 'application/octet-stream'

                main_type, sub_type = content_type.split('/', 1)
                with open(file, 'rb') as fp:
                    content = fp.read()

                if main_type == 'text':
                    msg = MIMEText(str(content), _subtype=sub_type)
                elif main_type == 'image':
                    msg = MIMEImage(content, _subtype=sub_type)
                elif main_type == 'audio':
                    msg = MIMEAudio(content, _subtype=sub_type)
                elif main_type == 'application':
                    msg = MIMEApplication(
                        content, _subtype=sub_type, _encoder=encode_base64
                    )
                else:
                    msg = MIMEBase(main_type, sub_type)
                    msg.set_payload(content)

                filename = os.path.basename(file)
                msg.add_header('Content-Disposition', 'attachment', filename=filename)
                message.attach(msg)

        service = self.get_service('gmail', 'v1')
        body = {'raw': base64.urlsafe_b64encode(message.as_bytes()).decode()}
        message = service.users().messages().send(userId='me', body=body).execute()

        return message

    @action
    def get_labels(self):
        """
        Returns the available labels on the GMail account
        """
        service = self.get_service('gmail', 'v1')
        results = service.users().labels().list(userId='me').execute()
        labels = results.get('labels', [])
        return labels


# vim:sw=4:ts=4:et:
